%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(erl_eval_SUITE).
-export([all/1]).

-export([guard_1/1, guard_2/1,
	 match_pattern/1,
	 match_bin/1,
	 string_plusplus/1,
	 pattern_expr/1,
         guard_3/1,
         lc/1,
         simple_cases/1,
         funs/1]).

%%
%% Define to run outside of test server
%%
%%-define(STANDALONE,1).

-import(lists,[concat/1, sort/1]).

-export([count_down/2, count_down_fun/0, do_apply/2, 
         local_func/3, local_func_value/2]).

-ifdef(STANDALONE).
-define(config(A,B),config(A,B)).
-export([config/2]).
-define(line, noop, ).
config(priv_dir,_) ->
    ".".
-else.
-include("test_server.hrl").
-export([init_per_testcase/2, fin_per_testcase/2]).
% Default timetrap timeout (set in init_per_testcase).
-define(default_timeout, ?t:minutes(1)).
init_per_testcase(_Case, Config) ->
    ?line Dog = ?t:timetrap(?default_timeout),
    [{watchdog, Dog} | Config].
fin_per_testcase(_Case, Config) ->
    Dog = ?config(watchdog, Config),
    test_server:timetrap_cancel(Dog),
    ok.
-endif.

all(doc) ->
    ["Test cases for the 'erl_eval' module."];
all(suite) ->
    [guard_1, guard_2, match_pattern, string_plusplus, pattern_expr,
     match_bin, guard_3, lc, simple_cases, funs].

guard_1(doc) ->
    ["(OTP-2405)"];
guard_1(suite) ->
    [];
guard_1(Config) when list(Config) ->
    ?line {ok,Tokens ,_} =
	erl_scan:string("if a+4 == 4 -> yes; true -> no end. "),
    ?line {ok, [Expr]} = erl_parse:parse_exprs(Tokens),
    ?line no = guard_1_compiled(),
    ?line {value, no, []} = erl_eval:expr(Expr, []),
    ok.

guard_1_compiled() ->
    if a+4 == 4 -> yes; true -> no end.

guard_2(doc) ->
    ["Similar to guard_1, but type-correct"];
guard_2(suite) ->
    [];
guard_2(Config) when list(Config) ->
    ?line {ok,Tokens ,_} =
	erl_scan:string("if 6+4 == 4 -> yes; true -> no end. "),
    ?line {ok, [Expr]} = erl_parse:parse_exprs(Tokens),
    ?line no = guard_2_compiled(),
    ?line {value, no, []} = erl_eval:expr(Expr, []),
    ok.

guard_2_compiled() ->
    if 6+4 == 4 -> yes; true -> no end.

string_plusplus(doc) ->
    ["OTP-3069: syntactic sugar string ++ ..."];
string_plusplus(suite) ->
    [];
string_plusplus(Config) when list(Config) ->
    ?line check(fun() -> case "abc" of "ab" ++ L -> L end end,
		"case \"abc\" of \"ab\" ++ L -> L end. ",
		"c"),
    ?line check(fun() -> case "abcde" of "ab" ++ "cd" ++ L -> L end end,
		"case \"abcde\" of \"ab\" ++ \"cd\" ++ L -> L end. ",
		"e"),
    ?line check(fun() -> case "abc" of [97, 98] ++ L -> L end end,
		"case \"abc\" of [97, 98] ++ L -> L end. ",
		"c"),
    ok.

match_pattern(doc) ->
    ["OTP-2983: match operator in pattern"];
match_pattern(suite) ->
    [];
match_pattern(Config) when list(Config) ->
    ?line check(fun() -> case {a, b} of {a, _X}=Y -> {x,Y} end end,
		"case {a, b} of {a, X}=Y -> {x,Y} end. ",
		{x, {a, b}}),
    ?line check(fun() -> case {a, b} of Y={a, _X} -> {x,Y} end end,
		"case {a, b} of Y={a, X} -> {x,Y} end. ",
		{x, {a, b}}),
    ?line check(fun() -> case {a, b} of Y={a, _X}=Z -> {Z,Y} end end,
		"case {a, b} of Y={a, X}=Z -> {Z,Y} end. ",
		{{a, b}, {a, b}}),
    ?line check(fun() -> A = 4, B = 28, <<13:(A+(X=B))>>, X end,
                "begin A = 4, B = 28, <<13:(A+(X=B))>>, X end.",
                28),
    ok.

match_bin(doc) ->
    ["binary match problems"];
match_bin(suite) ->
    [];
match_bin(Config) when list(Config) ->
    ?line check(fun() -> <<"abc">> = <<"abc">> end,
		"<<\"abc\">> = <<\"abc\">>. ",
		<<"abc">>),
    ?line check(fun() ->
			<<Size,B:Size/binary,Rest/binary>> = <<2,"AB","CD">>,
			{Size,B,Rest}
		end,
		"begin <<Size,B:Size/binary,Rest/binary>> = <<2,\"AB\",\"CD\">>, "
		"{Size,B,Rest} end. ",
		{2,<<"AB">>,<<"CD">>}),
    ok.

pattern_expr(doc) ->
    ["OTP-3144: compile-time expressions in pattern"];
pattern_expr(suite) ->
    [];
pattern_expr(Config) when list(Config) ->
    ?line check(fun() -> case 4 of 2+2 -> ok end end,
		"case 4 of 2+2 -> ok end. ",
		ok),
    ?line check(fun() -> case 2 of +2 -> ok end end,
		"case 2 of +2 -> ok end. ",
		ok),
    ok.

guard_3(doc) ->
    ["OTP-4518."];
guard_3(suite) ->
    [];
guard_3(Config) when list(Config) ->
    ?line check(fun() -> if false -> false; true -> true end end,
                "if false -> false; true -> true end.",
                true),
    ?line check(fun() -> if <<"hej">> == <<"hopp">> -> true; 
                            true -> false end end,
                "begin if <<\"hej\">> == <<\"hopp\">> -> true; 
                          true -> false end end.",
                false),
    ?line check(fun() -> if <<"hej">> == <<"hej">> -> true; 
                            true -> false end end,
                "begin if <<\"hej\">> == <<\"hej\">> -> true; 
                          true -> false end end.",
                true),
    ok.

lc(doc) ->
    ["OTP-4518."];
lc(suite) ->
    [];
lc(Config) when list(Config) ->
    ?line check(fun() -> X = 32, [X || X <- [1,2,3]] end,
                "begin X = 32, [X || X <- [1,2,3]] end.",
                [1,2,3]),
    ?line check(fun() -> X = 32, 
                         [X || <<X:X>> <- [<<1:32>>,<<2:32>>,<<3:8>>]] end,
    %% "binsize variable"          ^
                "begin X = 32, 
                 [X || <<X:X>> <- [<<1:32>>,<<2:32>>,<<3:8>>]] end.",
                [1,2]),
    ?line check(fun() -> Y = 13,[X || {X,Y} <- [{1,2}]] end,
                "begin Y = 13,[X || {X,Y} <- [{1,2}]] end.",
                [1]),
    ?line error_check("begin [A || X <- [{1,2}], 1 == A] end.",
                      {unbound_var,'A'}),
    ?line error_check("begin X = 32, 
                        [{Y,W} || X <- [1,2,32,Y=4], Z <- [1,2,W=3]] end.",
                      {unbound_var,'Y'}),
    ?line error_check("begin X = 32,<<A:B>> = <<100:X>> end.",
                      {unbound_var,'B'}),
    ?line check(fun() -> [X || X <- [1,2,3,4], not (X < 2)] end,
                "begin [X || X <- [1,2,3,4], not (X < 2)] end.",
                [2,3,4]),
    ?line check(fun() -> [X || X <- [true,false], X] end,
                "[X || X <- [true,false], X].", [true]),
    ok.

simple_cases(doc) ->
    ["Simple cases, just to cover some code."];
simple_cases(suite) ->
    [];
simple_cases(Config) when list(Config) ->
    ?line check(fun() -> A = $C end, "A = $C.", $C),
    %% ?line check(fun() -> A = 3.14 end, "A = 3.14.", 3.14),
    %% try...
    ?line check(fun() -> self() ! a, A = receive a -> true end end,
                "begin self() ! a, A = receive a -> true end end.",
                true),
    ?line check(fun() -> c:flush(), self() ! a, self() ! b, self() ! c,
                         receive b -> b end, 
                         {messages, [a,c]} ==
                             erlang:process_info(self(), messages),
                         c:flush() end,
                "begin c:flush(), self() ! a, self() ! b, self() ! c,"
                "receive b -> b end,"
                "{messages, [a,c]} =="
                "     erlang:process_info(self(), messages), c:flush() end.",
                ok),
    ?line check(fun() -> self() ! a, A = receive a -> true 
                                         after 0 -> false end end,
                "begin self() ! a, A = receive a -> true"
                "                      after 0 -> false end end.",
                true),
    ?line check(fun() -> c:flush(), self() ! a, self() ! b, self() ! c,
                         receive b -> b after 0 -> true end, 
                         {messages, [a,c]} ==
                             erlang:process_info(self(), messages),
                         c:flush() end,
                "begin c:flush(), self() ! a, self() ! b, self() ! c,"
                "receive b -> b after 0 -> true end,"
                "{messages, [a,c]} =="
                "     erlang:process_info(self(), messages), c:flush() end.",
                ok),
    ?line check(fun() -> receive _ -> true after 10 -> false end end,
                "receive _ -> true after 10 -> false end.",
                false),
    ?line check(fun() -> F = fun(A) -> A end, true = 3 == F(3) end,
                "begin F = fun(A) -> A end, true = 3 == F(3) end.",
                true),
    ?line check(fun() -> F = fun(A) -> A end, true = 3 == apply(F, [3]) end,
                "begin F = fun(A) -> A end, true = 3 == apply(F,[3]) end.",
                true),
    ?line check(fun() -> catch throw(a) end, "catch throw(a).", a),
    ?line check(fun() -> catch a end, "catch a.", a),
    ?line check(fun() -> 4 == 3 end, "4 == 3.", false),
    ?line check(fun() -> not true end, "not true.", false),
    ?line check(fun() -> -3 end, "-3.", -3),
    ?line check(fun() -> false andalso kludd end, "false andalso kludd.",
                false),
    ?line check(fun() -> true andalso true end, "true andalso true.",
                true),
    ?line check(fun() -> true andalso false end, "true andalso false.",
                false),
    ?line error_check("kladd andalso kludd.", {badarg,kladd}),
    ?line error_check("true andalso kludd.", {badarg,kludd}),
    ?line check(fun() -> true orelse kludd end, 
                "true orelse kludd.", true),
    ?line check(fun() -> false orelse false end, 
                "false orelse false.", false),
    ?line check(fun() -> false orelse true end, 
                "false orelse true.", true),
    ?line error_check("kladd orelse kludd.", {badarg,kladd}),
    ?line error_check("false orelse kludd.", {badarg,kludd}),
    ?line error_check("[X || X <- [1,2,3], erlang:process_info(self())].",
                      bad_filter),

    %% Just adopting to the R9-compiler: (filters are _not_ guards!)
    ?line check(fun() -> [X || X <- [1,2,3], X+2] end,
                "[X || X <- [1,2,3], X+2].", []),

    ?line check(fun() -> [X || X <- [1,2,3], [X] == [X || X <- [2]]] end,
                "[X || X <- [1,2,3], [X] == [X || X <- [2]]].",
                [2]),
    ?line check(fun() -> F = fun(1) -> ett; (2) -> zwei end, 
                         ett = F(1), zwei = F(2) end,
                "begin F = fun(1) -> ett; (2) -> zwei end, 
                         ett = F(1), zwei = F(2) end.",
                zwei),
    ?line check(fun() -> F = fun(X) when X == 1 -> ett; 
                                (X) when X == 2 -> zwei end, 
                         ett = F(1), zwei = F(2) end,
                "begin F = fun(X) when X == 1 -> ett; 
                              (X) when X == 2 -> zwei end, 
                         ett = F(1), zwei = F(2) end.",
                zwei),
    ?line error_check("begin F = fun(1) -> ett end, zwei = F(2) end.",
                      function_clause),
    ?line check(fun() -> if length([1]) == 1 -> yes; 
                            true -> no end end,
                "if length([1]) == 1 -> yes; 
                            true -> no end.",
                yes),
    ?line check(fun() -> if is_integer(3) -> true; true -> false end end,
                "if is_integer(3) -> true; true -> false end.", true),
    ?line check(fun() -> if integer(3) -> true; true -> false end end,
                "if integer(3) -> true; true -> false end.", true),
    ?line check(fun() -> if is_float(3) -> true; true -> false end end,
                "if is_float(3) -> true; true -> false end.", false),
    ?line check(fun() -> if float(3) -> true; true -> false end end,
                "if float(3) -> true; true -> false end.", false),
    ?line check(fun() -> if is_number(3) -> true; true -> false end end,
                "if is_number(3) -> true; true -> false end.", true),
    ?line check(fun() -> if number(3) -> true; true -> false end end,
                "if number(3) -> true; true -> false end.", true),
    ?line check(fun() -> if is_atom(a) -> true; true -> false end end,
                "if is_atom(a) -> true; true -> false end.", true),
    ?line check(fun() -> if atom(a) -> true; true -> false end end,
                "if atom(a) -> true; true -> false end.", true),
    ?line check(fun() -> if is_constant(a) -> true; true -> false end end,
                "if is_constant(a) -> true; true -> false end.", true),
    ?line check(fun() -> if constant(a) -> true; true -> false end end,
                "if constant(a) -> true; true -> false end.", true),
    ?line check(fun() -> if is_list([]) -> true; true -> false end end,
                "if is_list([]) -> true; true -> false end.", true),
    ?line check(fun() -> if list([]) -> true; true -> false end end,
                "if list([]) -> true; true -> false end.", true),
    ?line check(fun() -> if is_tuple({}) -> true; true -> false end end,
                "if is_tuple({}) -> true; true -> false end.", true),
    ?line check(fun() -> if tuple({}) -> true; true -> false end end,
                "if tuple({}) -> true; true -> false end.", true),
    ?line check(fun() -> if is_pid(self()) -> true; true -> false end end,
                "if is_pid(self()) -> true; true -> false end.", true),
    ?line check(fun() -> if pid(self()) -> true; true -> false end end,
                "if pid(self()) -> true; true -> false end.", true),
    ?line check(fun() -> R = make_ref(), if is_reference(R) -> true; 
                                            true -> false end end,
                "begin R = make_ref(), if is_reference(R) -> true;"
                "true -> false end end.", true),
    ?line check(fun() -> R = make_ref(), if reference(R) -> true; 
                                            true -> false end end,
                "begin R = make_ref(), if reference(R) -> true;"
                "true -> false end end.", true),
    ?line check(fun() -> if is_port(a) -> true; true -> false end end,
                "if is_port(a) -> true; true -> false end.", false),
    ?line check(fun() -> if port(a) -> true; true -> false end end,
                "if port(a) -> true; true -> false end.", false),
    ?line check(fun() -> if is_function(a) -> true; true -> false end end,
                "if is_function(a) -> true; true -> false end.", false),
    ?line check(fun() -> if function(a) -> true; true -> false end end,
                "if function(a) -> true; true -> false end.", false),
    ?line check(fun() -> if is_binary(<<>>) -> true; true -> false end end,
                "if is_binary(<<>>) -> true; true -> false end.", true),
    ?line check(fun() -> if binary(<<>>) -> true; true -> false end end,
                "if binary(<<>>) -> true; true -> false end.", true),
    ?line check(fun() -> if is_integer(a) == true -> yes; 
                            true -> no end end,
                "if is_integer(a) == true -> yes; 
                            true -> no end.",
                no),
    ?line check(fun() -> if [] -> true; true -> false end end,
                "if [] -> true; true -> false end.", false),
    ?line error_check("if lists:member(1,[1]) -> true; true -> false end.",
                      illegal_guard_expr),
    ?line error_check("if false -> true end.", if_clause),
    ?line check(fun() -> if a+b -> true; true -> false end end,
                "if a + b -> true; true -> false end.", false),
    ?line check(fun() -> if + b -> true; true -> false end end,
                "if + b -> true; true -> false end.", false),
    ?line error_check("case foo of bar -> true end.", {case_clause,foo}),
    ?line error_check("case 4 of 2+a -> true; _ -> false end.", 
                      illegal_pattern),
    ?line error_check("case 4 of +a -> true; _ -> false end.", 
                      illegal_pattern),
    ?line check(fun() -> case a of 
                             X when X == b -> one;
                             X when X == a -> two
                         end end,
                "begin case a of 
                             X when X == b -> one;
                             X when X == a -> two
                         end end.", two),
    ?line error_check("3 = 4.", {badmatch,4}),
    ?line error_check("a = 3.", {badmatch,3}),
    %% ?line error_check("3.1 = 2.7.",{badmatch,2.7}),
    ?line error_check("$c = 4.", {badmatch,4}),
    ?line check(fun() -> $c = $c end, "$c = $c.", $c),
    ?line check(fun() -> _ = bar end, "_ = bar.", bar),
    ?line check(fun() -> A = 14, A = 14 end, 
                "begin A = 14, A = 14 end.", 14),
    ?line error_check("begin A = 14, A = 16 end.", {badmatch,16}),
    ?line error_check("\"hej\" = \"san\".", {badmatch,"san"}),
    ?line check(fun() -> "hej" = "hej" end,
                "\"hej\" = \"hej\".", "hej"),
    ?line error_check("[] = [a].", {badmatch,[a]}),
    ?line check(fun() -> [] = [] end, "[] = [].", []),
    ?line error_check("[a] = [].", {badmatch,[]}),
    ?line error_check("{a,b} = 34.", {badmatch,34}),
    ?line error_check("<<8:7>>.", badarg),
    ?line error_check("<<34:32>> = \"hej\".", {badmatch,"hej"}),
    ?line check(fun() -> trunc((1 * 3 div 3 + 4 - 3) / 1) rem 2 end,
                "begin trunc((1 * 3 div 3 + 4 - 3) / 1) rem 2 end.", 0),
    ?line check(fun() -> (2#101 band 2#10101) bor (2#110 bxor 2#010) end,
                "(2#101 band 2#10101) bor (2#110 bxor 2#010).", 5),
    ?line check(fun() -> (2#1 bsl 4) + (2#10000 bsr 3) end,
                "(2#1 bsl 4) + (2#10000 bsr 3).", 18),
    ?line check(fun() -> ((1<3) and ((1 =:= 2) or (1 =/= 2))) xor (1=<2) end,
                "((1<3) and ((1 =:= 2) or (1 =/= 2))) xor (1=<2).", false),
    ?line check(fun() -> (a /= b) or (2 > 4) or (3 >= 3) end,
                "(a /= b) or (2 > 4) or (3 >= 3).", true),
    ?line check(fun() -> "hej" ++ "san" =/= "hejsan" -- "san" end,
                "\"hej\" ++ \"san\" =/= \"hejsan\" -- \"san\".", true),
    ?line check(fun() -> (bnot 1) < -0 end, "(bnot (+1)) < -0.", true),
    ok.

funs(doc) ->
    ["Simple cases, just to cover some code."];
funs(suite) ->
    [];
funs(Config) when list(Config) ->
    do_funs(none, none),
    do_funs(lfh(), none),
    do_funs(lfh(), efh()),

    ?line error_check("nix:foo().", {access_not_allowed,nix}, lfh(), efh()),
    ?line error_check("bar().", undef, none, none),

    ?line check(fun() -> F1 = fun(F,N) -> ?MODULE:count_down(F, N) end,
                         F1(F1, 1000) end,
                "begin F1 = fun(F,N) -> count_down(F, N) end,"
                "F1(F1,1000) end.",
		0, ['F1'], lfh(), none),

    ?line check(fun() -> F1 = fun(F,N) -> ?MODULE:count_down(F, N) end,
                         F1(F1, 1000) end,
                "begin F1 = fun(F,N) -> count_down(F, N) end,"
                "F1(F1,1000) end.",
		0, ['F1'], lfh_value(), none),

    ?line check(fun() -> F1 = fun(F,N) -> ?MODULE:count_down(F, N) end,
                         F1(F1, 1000) end,
                "begin F1 = fun(F,N) -> count_down(F, N) end,"
                "F1(F1,1000) end.",
		0, ['F1'], lfh_value_extra(), none),

    ?line check(fun() -> F1 = fun(F,N) -> ?MODULE:count_down(F, N) end,
                         F1(F1, 1000) end,
                "begin F1 = fun(F,N) -> count_down(F, N) end,"
                "F1(F1,1000) end.",
		0, ['F1'], {?MODULE,local_func_value}, none),
    %% This is not documented, and only for backward compatibility (good!).
    B0 = erl_eval:new_bindings(),
    ?line check(fun() -> is_function(?MODULE:count_down_fun()) end,
                "begin is_function(count_down_fun()) end.",
                true, [], {?MODULE,local_func,[B0]},none),

    ok.

do_funs(LFH, EFH) ->
    %% LFH is not really used by these examples...

    %% These tests do not prove that tail recursive functions really
    %% work (that the process does not grow); one should also run them
    %% manually with 1000 replaced by 1000000.

    M = atom_to_list(?MODULE),
    ?line check(fun() -> F1 = fun(F,N) -> ?MODULE:count_down(F, N) end,
                         F1(F1, 1000) end,
                concat(["begin F1 = fun(F,N) -> ", M, 
                        ":count_down(F, N) end, F1(F1,1000) end."]),
		0, ['F1'], LFH, EFH),
    ?line check(fun() -> F1 = fun(F,N) -> apply(?MODULE,count_down,[F,N]) 
                              end, F1(F1, 1000) end,
                concat(["begin F1 = fun(F,N) -> apply(", M, 
                        ",count_down,[F, N]) end, F1(F1,1000) end."]),
		0, ['F1'], LFH, EFH),
    ?line check(fun() -> F1 = fun(F,N) -> {?MODULE,count_down}(F,N)
                              end, F1(F1, 1000) end,
                concat(["begin F1 = fun(F,N) -> {", M, 
                        ",count_down}(F, N) end, F1(F1,1000) end."]),
		0, ['F1'], LFH, EFH),
    ?line check(fun() -> F = fun(F,N) when N > 0 -> apply(F,[F,N-1]); 
                                (_F,0) -> ok end, 
                         F(F, 1000)
                end,
                "begin F = fun(F,N) when N > 0 -> apply(F,[F,N-1]);"
                             "(_F,0) -> ok end,"
                       "F(F, 1000) end.",
                ok, ['F'], LFH, EFH),
    ?line check(fun() -> F = fun(F,N) when N > 0 -> 
                                     apply(erlang,apply,[F,[F,N-1]]);
                                (_F,0) -> ok end, 
                         F(F, 1000)
                end,
                "begin F = fun(F,N) when N > 0 ->"
                                   "apply(erlang,apply,[F,[F,N-1]]);"
                             "(_F,0) -> ok end,"
                       "F(F, 1000) end.",
                ok, ['F'], LFH, EFH),
    ?line check(fun() -> F = count_down_fun(), 
                         SF = fun(SF, F1, N) -> F(SF, F1, N) end,
                         SF(SF, F, 1000) end,
                concat(["begin F = ", M, ":count_down_fun(),"
                        "SF = fun(SF, F1, N) -> F(SF, F1, N) end,"
                        "SF(SF, F, 1000) end."]),
                ok, ['F','SF'], LFH, EFH),


    ?line check(fun() -> F = fun(X) -> A = 1+X, {X,A} end, 
                         true = {2,3} == F(2) end,
                "begin F = fun(X) -> A = 1+X, {X,A} end, 
                       true = {2,3} == F(2) end.", true, ['F'], LFH, EFH),
    ?line check(fun() -> F = fun(X) -> {erlang,'+'}(X,2) end, 
                         true = 3 == F(1) end,
                "begin F = fun(X) -> {erlang,'+'}(X,2) end," 
                "      true = 3 == F(1) end.", true, ['F'],
               LFH, EFH),
    ?line check(fun() -> F = fun(X) -> size(X) end,
                         ?MODULE:do_apply(F,<<"hej">>) end, 
                concat(["begin F = fun(X) -> size(X) end,",
                        M,":do_apply(F,<<\"hej\">>) end."]),
                3, ['F'], LFH, EFH),

    ?line check(fun() -> F1 = fun(X, Z) -> {X,Z} end,
                         Z = 5,
                         F2 = fun(X, Y) -> F1(Z,{X,Y}) end,
                         F3 = fun(X, Y) -> {a,F1(Z,{X,Y})} end,
                         {5,{x,y}} = F2(x,y), 
                         {a,{5,{y,x}}} = F3(y,x), 
                         {5,{5,y}} = F2(Z,y), 
                         true = {5,{x,5}} == F2(x,Z) end,
                "begin F1 = fun(X, Z) -> {X,Z} end,
                       Z = 5,
                       F2 = fun(X, Y) -> F1(Z,{X,Y}) end,
                       F3 = fun(X, Y) -> {a,F1(Z,{X,Y})} end,
                       {5,{x,y}} = F2(x,y), 
                       {a,{5,{y,x}}} = F3(y,x), 
                       {5,{5,y}} = F2(Z,y), 
                       true = {5,{x,5}} == F2(x,Z) end.",
                true, ['F1','Z','F2','F3'], LFH, EFH),
    ?line check(fun() -> F = fun(X) -> size(X) end,
                         F2 = fun(Y) -> F(Y) end,
                         ?MODULE:do_apply(F2,<<"hej">>) end, 
                concat(["begin F = fun(X) -> size(X) end,",
                        "F2 = fun(Y) -> F(Y) end,",
                        M,":do_apply(F2,<<\"hej\">>) end."]),
                3, ['F','F2'], LFH, EFH),
    ?line check(fun() -> Z = 5, F = fun(X) -> {Z,X} end,
                         F2 = fun(Z) -> F(Z) end, F2(3) end,
                "begin Z = 5, F = fun(X) -> {Z,X} end,
                       F2 = fun(Z) -> F(Z) end, F2(3) end.",
                {5,3},['F','F2','Z'], LFH, EFH),
    ?line check(fun() -> F = fun(Z) -> Z end,
                         F2 = fun(X) -> F(X), Z = {X,X}, Z end,
                         {1,1} = F2(1), Z = 7, Z end,
                "begin F = fun(Z) -> Z end,
                       F2 = fun(X) -> F(X), Z = {X,X}, Z end,
                       {1,1} = F2(1), Z = 7, Z end.", 7, ['F','F2','Z'], 
                LFH, EFH),
    ?line check(fun() -> F = fun(F, N) -> [?MODULE:count_down(F,N) || X <-[1]]
                             end, F(F,2) end,
                concat(["begin F = fun(F, N) -> [", M, 
                       ":count_down(F,N) || X <-[1]] end, F(F,2) end."]),
                [[[0]]], ['F'], LFH, EFH),
    
    ok.

count_down(F, N) when N > 0 ->
    F(F, N-1);
count_down(_F, N) ->
    N.

count_down_fun() ->
    fun(SF,F,N) when N > 0 -> SF(SF,F,N-1);
       (_SF,_F,_N) -> ok
    end.

do_apply(F, V) ->
    F(V).

lfh() ->
    {eval, fun(F, As, Bs) -> local_func(F, As, Bs) end}.

local_func(F, As0, Bs0) when atom(F) ->
    {As,Bs} = erl_eval:expr_list(As0, Bs0, {eval,lfh()}),
    case erlang:function_exported(?MODULE, F, length(As)) of
	true ->
	    {value,apply(?MODULE, F, As),Bs};
	false ->
	    {value,apply(shell_default, F, As),Bs}
    end.

lfh_value_extra() ->
    %% Not documented.
    {value, fun(F, As) -> local_func_value(F, As) end, []}.

lfh_value() ->
    {value, fun(F, As) -> local_func_value(F, As) end}.

local_func_value(F, As) when atom(F) ->
    case erlang:function_exported(?MODULE, F, length(As)) of
	true ->
	    apply(?MODULE, F, As);
	false ->
	    apply(shell_default, F, As)
    end.

efh() ->
    {value, fun(F, As) -> external_func(F, As) end}.

external_func({M,_}, _As) when M == nix ->
    exit({{access_not_allowed,M},[mfa]});
external_func(F, As) when function(F) ->
    apply(F, As);
external_func({M,F}, As) ->
    apply(M, F, As).

check(F, String, Result) ->
    ?line {ok,Tokens,_} =
	erl_scan:string(String),
    ?line {ok, [Expr]} = erl_parse:parse_exprs(Tokens),
    ?line Result = F(),
    ?line case catch erl_eval:expr(Expr, []) of
	      {value, Result, _} ->
		  ok;
	      Other ->
		  test_server:fail({eval, Other, Result})
	  end,
    ok.

check(F, String, Result, BoundVars, LFH, EFH) ->
    ?line {ok,Tokens,_} =
	erl_scan:string(String),
    ?line {ok, [Expr]} = erl_parse:parse_exprs(Tokens),
    ?line Result = F(),
    ?line case catch erl_eval:expr(Expr, [], LFH, EFH) of
	      {value, Result, Bs} ->
                  %% We just assume that Bs is an orddict...
                  Keys = orddict:fetch_keys(Bs),
                  case sort(BoundVars) == Keys of
                      true -> ok;
                      false -> 
                          test_server:fail({check, BoundVars, Keys})
                  end,
		  ok;
	      Other ->
		  test_server:fail({check, Other, Result})
	  end,
    ok.

error_check(String, Result) ->
    ?line {ok,Tokens,_} =
	erl_scan:string(String),
    ?line {ok, [Expr]} = erl_parse:parse_exprs(Tokens),
    ?line case catch erl_eval:expr(Expr, []) of
              {'EXIT', {Result,_}} ->
                  ok;
	      Other ->
		  test_server:fail({eval, Other, Result})
	  end,
    ok.

error_check(String, Result, LFH, EFH) ->
    ?line {ok,Tokens,_} =
	erl_scan:string(String),
    ?line {ok, [Expr]} = erl_parse:parse_exprs(Tokens),
    ?line case catch erl_eval:expr(Expr, [], LFH, EFH) of
              {'EXIT', {Result,_}} ->
                  ok;
	      Other ->
		  test_server:fail({eval, Other, Result})
	  end,
    ok.
