%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(ets_SUITE).
-export([all/1]).
-export([new/1,default/1,setbag/1,badnew/1,verybadnew/1,named/1,keypos2/1,
	 privacy/1,nosy/1]).
-export([insert/1,empty/1,badinsert/1]).
-export([lookup/1,time_lookup/1,badlookup/1]).
-export([delete/1,delete_elem/1,delete_tab/1,baddelete/1,match_delete/1]).
-export([match_delete3/1]).
-export([firstnext/1]).
-export([slot/1]).
-export([match/1, match1/1, match2/1, match_object/1, match_object2/1]).
-export([misc/1, dups/1, misc1/1, fixtable/1, info/1, tab2list/1]).
-export([files/1, tab2file/1, tab2file2/1, tab2file3/1]).
-export([heavy/1, heavy_lookup/1, heavy_lookup_element/1]).
-export([lookup_element/1, lookup_element_mult/1]).
-export([all_2214/1]).
%-export([i_2409/1]).
-export([fold/1]).
-export([foldl_ordered/1, foldr_ordered/1, foldl/1, foldr/1, fold_empty/1]).
-export([t_delete_object/1, t_init_table/1, t_whitebox/1, 
	 t_delete_all_objects/1, t_insert_list/1, t_test_ms/1,
	 t_select_delete/1,t_ets_dets/1]).

-export([do_lookup/2, do_lookup_element/3]).

-export([do0_2214/0, do_2214/1, do1_2214/1]).
-export([ordered/1, ordered_match/1, interface_equality/1,
	 fixtable_next/1, fixtable_insert/1, rename/1, update_counter/1,
	 partly_bound/1, match_heavy/1]).
-export([member/1]).
-export([memory/1]).
-export([select_fail/1]).
-export([init_per_testcase/2, fin_per_testcase/2]).
%% Convenience for manual testing
-export([random_test/0]).

% internal exports
%-export([heartbeat_loop/1]).
-export([dont_make_worse_sub/0, make_better_sub1/0, make_better_sub2/0]).

-include("test_server.hrl").

-changes(["- Added test case for match_delete.",
	  "- Added another match test case (specified keypos)",
          "- Improved misc test case a little.",
	  "- Added fixtable test case.",
	  "- Added doc clause to some test cases for test_server suite "
	  "structure generation (to be added)"
	 ]).
-changes(["- Added test cases for file2tab & tab2file w/ pretty heavy "
	  "loaded tables of both set & bag type."
	  "- Added heacy lookup & lookup_element tests"
	  "- Added ets:info test case."
	 ]).
-changes(["- Fixed all test cases (Config) clause to have list() "
	  "  guard"
	 ]).
-changes(["- Added case verybadnew OTP-2314 ets:new crashes emulator."
	 ]).

init_per_testcase(Func, Config) ->
    Dog=test_server:timetrap(test_server:seconds(360)),
    [{watchdog, Dog}|Config].

fin_per_testcase(Func, Config) ->
    Dog=?config(watchdog, Config),
    test_server:timetrap_cancel(Dog).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

all(suite) -> {req, [stdlib],
	       [new,insert,lookup,delete,firstnext,slot,match,
		all_2214,
		lookup_element, misc,files, heavy, 
%		i_2409, 
		ordered, ordered_match, interface_equality,
		fixtable_next, fixtable_insert, rename,
		update_counter, partly_bound, match_heavy,
		fold, member,
		t_delete_object, t_init_table, t_whitebox, 
		t_delete_all_objects, t_insert_list, t_test_ms,
		t_select_delete, t_ets_dets, memory,
		select_fail]}.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

new(suite) -> [default,setbag,badnew,verybadnew,named,keypos2,privacy].

default(doc) ->
    ["Test case to check that a new ets table is defined as a `set' and "
     "`protected'"];
default(suite) -> [];
default(Config) when list(Config) ->
    %% Default should be set,protected
    ?line Def = ets:new(def,[]),
    ?line set = ets:info(Def,type),
    ?line protected = ets:info(Def,protection),
    ok.

select_fail(doc) ->
    ["Test that select fails even if nothing can match"];
select_fail(suite) ->
    [];
select_fail(Config) when list(Config) ->    
    ?line lists:foreach(
	    fun(T) ->
		    ?line ets:insert(T,{a,a}),
		    ?line case (catch 
				ets:select(T,[{{a,'_'},[],[{snuffla}]}])) of
			      {'EXIT',{badarg,_}} ->
				  ok;
			      Else0 ->
				  exit({type,ets:info(T,type),
					expected,'EXIT',got,Else0})
			  end,
		    ?line case (catch 
				ets:select(T,[{{b,'_'},[],[{snuffla}]}])) of
			      {'EXIT',{badarg,_}} ->
				  ok;
			      Else1 ->
				  exit({type,ets:info(T,type),
					expected,'EXIT',got,Else1})
			  end,
		    ?line ets:delete(T)
	    end, [ets:new(x,[ordered_set]),ets:new(x,[set]),
		  ets:new(x,[bag]),ets:new(x,[duplicate_bag])]),
    ok.
 

-define(S(T),ets:info(T,memory)).
memory(doc) ->
    ["Whitebox test of ets:info(X,memory)"];
memory(suite) ->
    [];
memory(Config) when list(Config) ->    
    ?line L = [T1,T2,T3,T4] = fill_sets_int(1000),
    ?line {17856,17063,17063,17069} = {?S(T1),?S(T2),?S(T3),?S(T4)},
    ?line lists:foreach(fun(T) ->
			  ?line ets:delete(T,ets:first(T))
		  end,
		  L),
    ?line {17842,17046,17029,17035} ={?S(T1),?S(T2),?S(T3),?S(T4)},
    ?line lists:foreach(fun(T) ->
			  ?line ets:match_delete(T,{ets:first(T),'_'})
		  end,
		  L),
    ?line {17828,17031,16997,17001} = {?S(T1),?S(T2),?S(T3),?S(T4)},
    ?line lists:foreach(fun(T) ->
			  ?line ets:delete_all_objects(T)
		  end,
		  L),
    ?line {70,277,277,277} = {?S(T1),?S(T2),?S(T3),?S(T4)},
    lists:foreach(fun(T) ->
			  ?line ets:delete(T)
		  end,
		  L),
    ?line L2 =  [T11,T12,T13,T14] = fill_sets_int(1000),
    ?line lists:foreach(fun(T) ->
			  ?line ets:select_delete(T,[{'_',[],[true]}])
		  end,
		  L2),
    ?line {70,277,277,277} = {?S(T11),?S(T12),?S(T13),?S(T14)},
    ok.

t_whitebox(doc) ->
    ["Diverse whitebox testes"];
t_whitebox(suite) ->
    [];
t_whitebox(Config) when list(Config) ->    
    whitebox_1(),
    whitebox_1(),
    whitebox_1(),
    whitebox_2(),
    whitebox_2(),
    whitebox_2().

whitebox_1() ->
    ?line T=ets:new(x,[bag]),
    ?line ets:insert(T,[{du,glade},{ta,en}]),
    ?line ets:insert(T,[{hej,hopp2},{du,glade2},{ta,en2}]),
    ?line {_,C}=ets:match(T,{ta,'$1'},1),
    ?line ets:select(C),
    ?line ets:match(C),
    ?line ets:delete(T),
    ok.

whitebox_2() ->
    ?line T=ets:new(x,[ordered_set,{keypos,2}]),
    ?line T2=ets:new(x,[set,{keypos,2}]),
    ?line 0 = ets:select_delete(T,[{{hej},[],[true]}]),
    ?line 0 = ets:select_delete(T,[{{hej,hopp},[],[true]}]),
    ?line 0 = ets:select_delete(T2,[{{hej},[],[true]}]),
    ?line 0 = ets:select_delete(T2,[{{hej,hopp},[],[true]}]),
    ?line ets:delete(T),
    ?line ets:delete(T2),
    ok.
    
    
t_ets_dets(doc) ->
    ["Test ets:to/from_dets"];
t_ets_dets(suite) ->
    [];
t_ets_dets(Config) when list(Config) ->
    ?line Fname = gen_dets_filename(Config,1),
    ?line (catch file:delete(Fname)),
    ?line {ok,DTab} = dets:open_file(testdets_1,
			       [{file, Fname}]),
    ?line ETab = ets:new(x,[]),
    ?line filltabint(ETab,3000),
    ?line DTab = ets:to_dets(ETab,DTab),
    ?line ets:delete_all_objects(ETab),
    ?line 0 = ets:info(ETab,size),
    ?line true = ets:from_dets(ETab,DTab),
    ?line 3000 = ets:info(ETab,size),
    ?line ets:delete(ETab),
    ?line {'EXIT',{badarg,[{ets,to_dets,[ETab,DTab]}|_]}} =
	(catch ets:to_dets(ETab,DTab)),
    ?line {'EXIT',{badarg,[{ets,from_dets,[ETab,DTab]}|_]}} =
	(catch ets:from_dets(ETab,DTab)),
    ?line ETab2 = ets:new(x,[]),
    ?line filltabint(ETab2,3000),
    ?line dets:close(DTab),
    ?line {'EXIT',{badarg,[{ets,to_dets,[ETab2,DTab]}|_]}} =
	(catch ets:to_dets(ETab2,DTab)),
    ?line {'EXIT',{badarg,[{ets,from_dets,[ETab2,DTab]}|_]}} =
	(catch ets:from_dets(ETab2,DTab)),
    ?line ets:delete(ETab2),
    ok.

t_delete_all_objects(doc) ->
    ["Test ets:delete_all_objects/1"];
t_delete_all_objects(suite) ->
    [];
t_delete_all_objects(Config) when list(Config) ->
    ?line T=ets:new(x,[]),
    ?line filltabint(T,4000),
    ?line O=ets:first(T),
    ?line ets:next(T,O),
    ?line ets:safe_fixtable(T,true),
    ?line true = ets:delete_all_objects(T),
    ?line '$end_of_table' = ets:next(T,O),
    ?line 0 = ets:info(T,size),
    ?line 4000 = ets:info(T,kept_objects),
    ?line ets:safe_fixtable(T,false),
    ?line 0 = ets:info(T,size),
    ?line 0 = ets:info(T,kept_objects),
    ?line filltabint(T,4000),
    ?line 4000 = ets:info(T,size),
    ?line true = ets:delete_all_objects(T),
    ?line 0 = ets:info(T,size),
    ?line ets:delete(T),
    ok.

t_delete_object(doc) ->
    ["Test ets:delete_object/2"];
t_delete_object(suite) ->
    [];
t_delete_object(Config) when list(Config) ->
    ?line T = ets:new(x,[]),
    ?line filltabint(T,4000),
    ?line del_one_by_one_set(T,1,4001),
    ?line filltabint(T,4000),
    ?line del_one_by_one_set(T,4000,0),
    ?line filltabint(T,4000),
    ?line First = ets:first(T),
    ?line Next = ets:next(T,First),
    ?line ets:safe_fixtable(T,true),
    ?line ets:delete_object(T,{First, integer_to_list(First)}),
    ?line Next = ets:next(T,First),
    ?line 3999 = ets:info(T,size),
    ?line 1 = ets:info(T,kept_objects),
    ?line ets:safe_fixtable(T,false),
    ?line 3999 = ets:info(T,size),
    ?line 0 = ets:info(T,kept_objects),
    ?line ets:delete(T),
    ?line T1 = ets:new(x,[ordered_set]),
    ?line filltabint(T1,4000),
    ?line del_one_by_one_set(T1,1,4001),
    ?line filltabint(T1,4000),
    ?line del_one_by_one_set(T1,4000,0),
    ?line ets:delete(T1),
    ?line T2 = ets:new(x,[bag]),
    ?line filltabint2(T2,4000),
    ?line del_one_by_one_bag(T2,1,4001),
    ?line filltabint2(T2,4000),
    ?line del_one_by_one_bag(T2,4000,0),
    ?line ets:delete(T2),
    ?line T3 = ets:new(x,[duplicate_bag]),
    ?line filltabint3(T3,4000),
    ?line del_one_by_one_dbag_1(T3,1,4001),
    ?line filltabint3(T3,4000),
    ?line del_one_by_one_dbag_1(T3,4000,0),
    ?line filltabint(T3,4000),
    ?line filltabint3(T3,4000),
    ?line del_one_by_one_dbag_2(T3,1,4001),
    ?line filltabint(T3,4000),
    ?line filltabint3(T3,4000),
    ?line del_one_by_one_dbag_2(T3,4000,0),
    ?line ets:delete(T3),
    ok.

make_init_fun(N) when N > 4000->
    fun(read) ->
	    end_of_input;
       (close) ->
	    exit(close_not_expected)
    end;
make_init_fun(N) ->
    fun(read) ->
	    case N rem 2 of
		0 ->
		    {[{N, integer_to_list(N)}, {N, integer_to_list(N)}],
		     make_init_fun(N + 1)};
		1 ->
		    {[], make_init_fun(N + 1)}
	    end;
       (close) ->
	    exit(close_not_expected)
    end.

t_init_table(doc) ->
    ["Test ets:init_table/2"];
t_init_table(suite) ->
    [];
t_init_table(Config) when list(Config)->
    ?line T = ets:new(x,[duplicate_bag]),
    ?line filltabint(T,4000),
    ?line ets:init_table(T, make_init_fun(1)),
    ?line del_one_by_one_dbag_1(T,4000,0),
    ?line ets:delete(T),
    ok.

do_fill_dbag_using_lists(T,0) ->
    T;
do_fill_dbag_using_lists(T,N) ->
    ets:insert(T,[{N,integer_to_list(N)},
		  {N + N rem 2,integer_to_list(N + N rem 2)}]),
    do_fill_dbag_using_lists(T,N - 1). 

t_insert_list(doc) ->
    ["Test ets:insert/2 with list of objects."];
t_insert_list(suite) ->
    [];
t_insert_list(Config) when list(Config) ->
    ?line T = ets:new(x,[duplicate_bag]),
    ?line do_fill_dbag_using_lists(T,4000),
    ?line del_one_by_one_dbag_2(T,4000,0),
    ?line ets:delete(T),
    ok.

t_test_ms(doc) ->
    ["Test interface of ets:test_ms/2"];
t_test_ms(suite) ->
    [];
t_test_ms(Config) when list(Config) ->
    ?line {ok,[a,b]} = ets:test_ms({a,b},
				   [{{'$1','$2'},[{'<','$1','$2'}],['$$']}]),
    ?line {ok,false} = ets:test_ms({a,b},
				   [{{'$1','$2'},[{'>','$1','$2'}],['$$']}]),
    ?line {error,[{error,String}]} = ets:test_ms({a,b},
						 [{{'$1','$2'},
						   [{'flurp','$1','$2'}],
						   ['$$']}]),
    ?line true = (if list(String) -> true; true -> false end),
    ok.

t_select_delete(doc) ->
    ["Test the ets:select_delete/2 and ets:select_count/2 BIF's"];
t_select_delete(suite) ->
    [];
t_select_delete(Config) when list(Config) ->
    ?line Tables = fill_sets_int(10000),
    lists:foreach
      (fun(Table) ->
	       ?line 4000 = ets:select_count(Table,[{{'$1', '_'},
						     [{'>', 
						       {'rem', 
							'$1', 5}, 
						       2}],
						     [true]}]),
	       ?line 4000 = ets:select_delete(Table,[{{'$1', '_'},
						      [{'>', 
							{'rem', 
							 '$1', 5}, 
							2}],
						      [true]}]),
	       ?line check(Table,
			   fun({N,_}) when (N rem 5) =< 2 ->
				   true;
			      (_) ->
				   false
			   end,
			   6000)
       end,
       Tables),
    lists:foreach
      (fun(Table) ->
	       ?line ets:select_delete(Table,[{'_',[],[true]}]),
	       ?line xfilltabint(Table,4000),
	       ?line successive_delete(Table,1,4001,bound),
	       ?line 0 = ets:info(Table,size),
	       ?line xfilltabint(Table,4000),
	       ?line successive_delete(Table,4000,0, bound),
	       ?line 0 = ets:info(Table,size),
	       ?line xfilltabint(Table,4000),
	       ?line successive_delete(Table,1,4001,unbound),
	       ?line 0 = ets:info(Table,size),
	       ?line xfilltabint(Table,4000),
	       ?line successive_delete(Table,4000,0, unbound),
	       ?line 0 = ets:info(Table,size)
       end,
       Tables),
    lists:foreach
      (fun(Table) ->
	       F = case ets:info(Table,type) of
		       X when X == bag; X == duplicate_bag ->
			   2;
		       _ -> 
			   1
		   end,
 	       ?line xfilltabstr(Table, 4000),
	       ?line 1000 = ets:select_count(Table,
					     [{{[$3 | '$1'], '_'},
					       [{'==',
						 {'length', '$1'},
						 3}],[true]}]) div F,
	       ?line 1000 = ets:select_delete(Table,
					      [{{[$3 | '$1'], '_'},
						[{'==',
						  {'length', '$1'},
						  3}],[true]}]) div F,
	       ?line check(Table, fun({[3,_,_,_],_}) -> false;
				     (_) -> true
				  end, 3000*F),
	       ?line 8 = ets:select_count(Table,
					  [{{"7",'_'},[],[false]},
					   {{['_'], '_'},
					    [],[true]}]) div F,
	       ?line 8 = ets:select_delete(Table,
					   [{{"7",'_'},[],[false]},
					    {{['_'], '_'},
					     [],[true]}]) div F,
	       ?line check(Table, fun({"7",_}) -> true;
				     ({[_],_}) -> false;
				     (_) -> true
				  end, 2992*F),
	       ?line xfilltabstr(Table, 4000),
	       %% This happens to be interesting for other select types too
	       ?line 200 = length(ets:select(Table,
					     [{{[$3,'_','_'],'_'},
					       [],[true]},
					      {{[$1,'_','_'],'_'},
					       [],[true]}])) div F,
	       ?line 200 = ets:select_count(Table,
					    [{{[$3,'_','_'],'_'},
					      [],[true]},
					     {{[$1,'_','_'],'_'},
					      [],[true]}]) div F,
	       ?line 200 = length(element(1,ets:select(Table,
						       [{{[$3,'_','_'],'_'},
							 [],[true]},
							{{[$1,'_','_'],'_'},
							 [],[true]}],
						       1000))) div F,
	       ?line 200 = length(
			     ets:select_reverse(Table,
						[{{[$3,'_','_'],'_'},
						  [],[true]},
						 {{[$1,'_','_'],'_'},
						  [],[true]}])) div F,
	       ?line 200 = length(
			     element(1,
				     ets:select_reverse
				     (Table,
				      [{{[$3,'_','_'],'_'},
					[],[true]},
				       {{[$1,'_','_'],'_'},
					[],[true]}],
				      1000))) div F,
	       ?line 200 = ets:select_delete(Table,
					     [{{[$3,'_','_'],'_'},
					       [],[true]},
					      {{[$1,'_','_'],'_'},
					       [],[true]}]) div F,
	       ?line 0 = ets:select_count(Table,
					  [{{[$3,'_','_'],'_'},
					    [],[true]},
					   {{[$1,'_','_'],'_'},
					    [],[true]}]) div F,
	       ?line check(Table, fun({[$3,_,_],_}) -> false;
				     ({[$1,_,_],_}) -> false;
				     (_) -> true
				  end, 3800*F)
       end,
       Tables),
    lists:foreach(fun(Tab) -> ets:delete(Tab) end,Tables),
    ok.

partly_bound(doc) ->
    ["Test that partly bound keys gives faster matches"];
partly_bound(suite) ->
    [];
partly_bound(Config) when list(Config) ->
    ?line dont_make_worse(),
    ?line make_better(),
    ok.

dont_make_worse() ->
    seventyfive_percent_success({?MODULE,dont_make_worse_sub,[]},0,0,10).

dont_make_worse_sub() ->
    ?line T = build_table([a,b],[a,b],15000),
    ?line T1 = time_match_object(T,{'_',a,a,1500}, [{{a,a,1500},a,a,1500}]),
    ?line T2 = time_match_object(T,{{a,a,'_'},a,a,1500}, 
				 [{{a,a,1500},a,a,1500}]),
    ?line ets:delete(T),
    ?line true = (T1 > T2),
    ok.
    
make_better() ->
    seventyfive_percent_success({?MODULE,make_better_sub2,[]},0,0,10),
    seventyfive_percent_success({?MODULE,make_better_sub1,[]},0,0,10).
make_better_sub1() ->
    ?line T = build_table2([a,b],[a,b],15000),
    ?line T1 = time_match_object(T,{'_',1500,a,a}, [{{1500,a,a},1500,a,a}]),
    ?line T2 = time_match_object(T,{{1500,a,'_'},1500,a,a}, 
				 [{{1500,a,a},1500,a,a}]),
    ?line ets:delete(T),
    ?line io:format("~p>~p~n",[(T1 / 100),T2]),
    ?line true = ((T1 / 100) > T2), % More marginal than needed.
    ok.

make_better_sub2() ->
    ?line T = build_table2([a,b],[a,b],15000),
    ?line T1 = time_match(T,{'$1',1500,a,a}),
    ?line T2 = time_match(T,{{1500,a,'$1'},1500,a,a}),
    ?line ets:delete(T),
    ?line io:format("~p>~p~n",[(T1 / 100),T2]),
    ?line true = ((T1 / 100) > T2), % More marginal than needed.
    ok.


match_heavy(doc) ->
    ["Heavy random matching, comparing set with ordered_set."];
match_heavy(suite) ->
    [];
match_heavy(Config) when list(Config) ->
    PrivDir = ?config(priv_dir,Config),
    DataDir = ?config(data_dir, Config),
    %% Easier to have in process dictionary when manually
    %% running the test function.
    put(where_to_read,DataDir),
    put(where_to_write,PrivDir),
    Dog=?config(watchdog, Config),
    test_server:timetrap_cancel(Dog),    
    NewDog=test_server:timetrap(test_server:seconds(1000)),
    NewConfig = [{watchdog, NewDog} | lists:keydelete(watchdog,1,Config)],
    random_test(),
    drop_match(),
    NewConfig.

%%% Extra safety for the very low probability that this is not
%%% caught by the random test (Statistically impossible???) 
drop_match() ->
    ?line T = build_table([a,b],[a],1500),
    ?line [{{a,a,1},a,a,1},{{b,a,1},b,a,1}] = 
	ets:match_object(T, {'_','_','_',1}).


ets_match(Tab,Expr) ->
    case random:uniform(2) of
	1 ->
	    ets:match(Tab,Expr);
	_ ->
	    match_chunked(Tab,Expr)
    end.

match_chunked(Tab,Expr) ->
    match_chunked_collect(ets:match(Tab,Expr,
				    random:uniform(1999) + 1)).
match_chunked_collect('$end_of_table') ->
    [];
match_chunked_collect({Results, Continuation}) ->
    Results ++ match_chunked_collect(ets:match(Continuation)).

ets_match_object(Tab,Expr) ->
    case random:uniform(2) of
	1 ->
	    ets:match_object(Tab,Expr);
	_ ->
	    match_object_chunked(Tab,Expr)
    end.

match_object_chunked(Tab,Expr) ->
    match_object_chunked_collect(ets:match_object(Tab,Expr,
						  random:uniform(1999) + 1)).
match_object_chunked_collect('$end_of_table') ->
    [];
match_object_chunked_collect({Results, Continuation}) ->
    Results ++ match_object_chunked_collect(ets:match_object(Continuation)).


    
random_test() ->
    ?line ReadDir = get(where_to_read),
    ?line WriteDir = get(where_to_write),
    ?line (catch file:make_dir(WriteDir)),
    ?line Seed = case file:consult(filename:join([ReadDir, 
					    "preset_random_seed.txt"])) of
	       {ok,[X]} ->
		   X;
	       _ ->
		   {A,B,C} = erlang:now(),
		   random:seed(A,B,C),
		   get(random_seed)
	   end,
    put(random_seed,Seed),
    ?line {ok, F} = file:open(filename:join([WriteDir, 
					     "last_random_seed.txt"]), 
			      [write]),
    io:format(F,"~p. ~n",[Seed]),
    file:close(F),
    io:format("Random seed ~p written to ~s, copy to ~s to rerun with "
	      "same seed.",[Seed, 
			    filename:join([WriteDir, "last_random_seed.txt"]),
			    filename:join([ReadDir, 
					    "preset_random_seed.txt"])]),
    do_random_test().

do_random_test() ->
    ?line OrdSet = ets:new(xxx,[ordered_set]),
    ?line Set = ets:new(xxx,[]),
    ?line do_n_times(fun() ->
		       ?line Key = create_random_string(25),
		       ?line Value = create_random_tuple(25),
		       ?line ets:insert(OrdSet,{Key,Value}),
		       ?line ets:insert(Set,{Key,Value})
	       end, 5000),
    ?line io:format("~nData inserted~n"),
    ?line do_n_times(fun() ->
		       ?line I = random:uniform(25),
		       ?line Key = create_random_string(I) ++ '_',
		       ?line L1 = ets_match_object(OrdSet,{Key,'_'}),
		       ?line L2 = lists:sort(ets_match_object(Set,{Key,'_'})),
		       case L1 == L2 of
			   false ->
			       io:format("~p != ~p~n",
					 [L1,L2]),
			       ?line exit({not_eq, L1, L2});
			   true ->
			       ok
		       end
	       end,
	       2000),
    ?line io:format("~nData matched~n"),
    ?line ets:match_delete(OrdSet,'_'),
    ?line ets:match_delete(Set,'_'),
    ?line do_n_times(fun() ->
		       ?line Value = create_random_string(25),
		       ?line Key = create_random_tuple(25),
		       ?line ets:insert(OrdSet,{Key,Value}),
		       ?line ets:insert(Set,{Key,Value})
	       end, 2000),
    ?line io:format("~nData inserted~n"),
    (fun() ->
	     ?line Key = list_to_tuple(lists:duplicate(25,'_')),
	     ?line L1 = ets_match_object(OrdSet,{Key,'_'}),
	     ?line L2 = lists:sort(ets_match_object(Set,{Key,'_'})),
	     ?line 2000 = length(L1),
	     case L1 == L2 of
		 false ->
		     io:format("~p != ~p~n",
			       [L1,L2]),
		     ?line exit({not_eq, L1, L2});
		 true ->
		     ok
	     end
     end)(),
    (fun() ->
	     ?line Key = {'$1','$2','$3','$4',
			  '$5','$6','$7','$8',
			  '$9','$10','$11','$12',
			  '$13','$14','$15','$16',
			  '$17','$18','$19','$20',
			  '$21','$22','$23','$24',
			  '$25'},
	     ?line L1 = ets_match_object(OrdSet,{Key,'_'}),
	     ?line L2 = lists:sort(ets_match_object(Set,{Key,'_'})),
	     ?line 2000 = length(L1),
	     case L1 == L2 of
		 false ->
		     io:format("~p != ~p~n",
			       [L1,L2]),
		     ?line exit({not_eq, L1, L2});
		 true ->
		     ok
	     end
     end)(),
    (fun() ->
	     ?line Key = {'$1','$2','$3','$4',
			  '$5','$6','$7','$8',
			  '$9','$10','$11','$12',
			  '$13','$14','$15','$16',
			  '$17','$18','$19','$20',
			  '$21','$22','$23','$24',
			  '$25'},
	     ?line L1 = ets_match(OrdSet,{Key,'_'}),
	     ?line L2 = lists:sort(ets_match(Set,{Key,'_'})),
	     ?line 2000 = length(L1),
	     case L1 == L2 of
		 false ->
		     io:format("~p != ~p~n",
			       [L1,L2]),
		     ?line exit({not_eq, L1, L2});
		 true ->
		     ok
	     end
     end)(),
    ?line ets:match_delete(OrdSet,'_'),
    ?line ets:match_delete(Set,'_'),
    ?line do_n_times(fun() ->
		       ?line Value = create_random_string(25),
		       ?line Key = create_random_tuple(25),
		       ?line ets:insert(OrdSet,{Key,Value}),
		       ?line ets:insert(Set,{Key,Value})
	       end, 2000),
    ?line io:format("~nData inserted~n"),
    do_n_times(fun() ->
		       ?line Key = create_partly_bound_tuple(25),
		       ?line L1 = ets_match_object(OrdSet,{Key,'_'}),
		       ?line L2 = lists:sort(ets_match_object(Set,{Key,'_'})),
		       case L1 == L2 of
			   false ->
			       io:format("~p != ~p~n",
					 [L1,L2]),
			       ?line exit({not_eq, L1, L2});
			   true ->
			       ok
		       end
	       end,
	       2000),
    ?line do_n_times(fun() ->
		       ?line Key = create_partly_bound_tuple2(25),
		       ?line L1 = ets_match_object(OrdSet,{Key,'_'}),
		       ?line L2 = lists:sort(ets_match_object(Set,{Key,'_'})),
		       case L1 == L2 of
			   false ->
			       io:format("~p != ~p~n",
					 [L1,L2]),
			       ?line exit({not_eq, L1, L2});
			   true ->
			       ok
		       end
	       end,
	       2000),
    do_n_times(fun() ->
		       ?line Key = create_partly_bound_tuple2(25),
		       ?line L1 = ets_match(OrdSet,{Key,'_'}),
		       ?line L2 = lists:sort(ets_match(Set,{Key,'_'})),
		       case L1 == L2 of
			   false ->
			       io:format("~p != ~p~n",
					 [L1,L2]),
			       ?line exit({not_eq, L1, L2});
			   true ->
			       ok
		       end
	       end,
	       2000),
    io:format("~nData matched~n"),
    ets:match_delete(OrdSet,'_'),
    ets:match_delete(Set,'_'),
    do_n_times(fun() ->
		       do_n_times(fun() ->
					  ?line Value = 
					      create_random_string(25),
					  ?line Key = create_random_tuple(25),
					  ?line ets:insert(OrdSet,{Key,Value}),
					  ?line ets:insert(Set,{Key,Value})
				  end, 500),
		       io:format("~nData inserted~n"),
		       do_n_times(fun() ->
					  ?line Key = 
					      create_partly_bound_tuple(25),
					  ets:match_delete(OrdSet,{Key,'_'}),
					  ets:match_delete(Set,{Key,'_'}),
					  L1 = ets:info(OrdSet,size),
					  L2 = ets:info(Set,size),
					  [] = ets_match_object(OrdSet,
								{Key,'_'}),
					  case L1 == L2 of
					      false ->
						  io:format("~p != ~p "
							    "(deleted ~p)~n",
							    [L1,L2,Key]),
						  exit({not_eq, L1, L2,
							{deleted,Key}});
					      true ->
						  ok
					  end
				  end,
				  50),
		       io:format("~nData deleted~n")
	       end,
	       10),
    ets:delete(OrdSet),
    ets:delete(Set),
    ok.

update_counter(doc) ->
    ["test various variants of update_counter"];
update_counter(suite) ->
    [];
update_counter(Config) when list(Config) ->
    Set = ets:new(set,[]),
    OrdSet = ets:new(ordered_set,[ordered_set]),
    update_counter(Set,Config),
    update_counter(OrdSet, Config),
    ok.

update_counter(T,C) ->
    ?line ets:insert(T,{a,1,1}),
    ?line 101 = ets:update_counter(T,a,100),
    ?line [{a,101,1}] = ets:lookup(T,a),
    ?line 101 = ets:update_counter(T,a,{3,100}),
    ?line [{a,101,101}] = ets:lookup(T,a),
    ?line LoopF = fun(Initial, Times, Add, Tab, Key, Pos, Myself) ->
		    case Times of
			0 ->
			    Initial;
			N ->
			    NInitial = Initial + Add,
			    ?line NInitial = ets:update_counter(Tab,Key,
								{Pos,Add}),
			    ?line Myself(NInitial,N-1,Add,Tab,Key,Pos,Myself)
		    end
	    end,
    ?line New2 =LoopF(101,100,10000000000,T,a,2,LoopF),
    ?line New3 = LoopF(101,100,10000000000,T,a,3,LoopF),
    ?line LoopF(New2,200,-20000000000,T,a,2,LoopF),
    ?line LoopF(New3,20000,-20000,T,a,3,LoopF),
    ?line LoopF2 = fun(Initial, Times, Add, Tab, Key, 
		       Pos, Thres, Warp, Myself) ->
		    case Times of
			0 ->
			    Initial;
			N ->
			    NInitial = case Initial + Add of
					   X when X > Thres, Add > 0 ->
					       Warp;
					   Y when Y < Thres, Add < 0 ->
					       Warp;
					   Z ->
					       Z
				     end,
			    %erlang:display({Initial,NInitial,Add,Thres}),
			    ?line NInitial = ets:update_counter(Tab,Key,
								{Pos,Add,
								 Thres,Warp}),
			    ?line Myself(NInitial,N-1,Add,Tab,Key,Pos,
					 Thres, Warp, Myself)
		    end
	    end,
    ?line ets:insert(T,{a,101,101}),
    ?line New4 =LoopF2(101,100,10000000000,T,a,2,100000000000,101,LoopF2),
    ?line New5 = LoopF2(101,100,10000000000,T,a,3,100000000000,101,LoopF2),
    ?line LoopF2(New4,200,-20000000000,T,a,2,-200000000000,New4,LoopF2),
    ?line LoopF2(New5,20000,-20000,T,a,3,-200000,New5+1,LoopF2),
    %?line ets:insert(T,{b,4712,50000}),
    %?line 16#FFFFFFFF = ets:update_counter(T,b,{2,0,0,16#FFFFFFFF}),
    %?line -16#FFFFFFFF = ets:update_counter(T,b,{3,0,0,-16#FFFFFFFF}),
    ok.
		    


fixtable_next(doc) ->
    ["Check that a first-next sequence always works on a fixed table"];
fixtable_next(suite) ->
    [];
fixtable_next(Config) when list(Config) ->
    ?line do_fixtable_next(ets:new(set,[public])),
    ?line do_fixtable_next(ets:new(ordered_set, [ordered_set,public])),
    ?line do_fixtable_next(ets:new(bag, [bag,public])),
    ?line do_fixtable_next(ets:new(duplicate_bag, [duplicate_bag,public])),
    ok.
    
do_fixtable_next(Tab) ->
   ?line  F = fun(X,T,FF) -> case X of 
			   0 -> true; 
			   _ -> 
			       ets:insert(T, {X,
					      integer_to_list(X),
					      X rem 10}),
			       FF(X-1,T,FF) 
		       end 
	end,
    ?line F(100,Tab,F),
    ?line ets:fixtable(Tab,true),
    ?line First = ets:first(Tab),
    ?line ets:delete(Tab, First),
    ?line ets:next(Tab, First),
    ?line ets:match_delete(Tab,{'_','_','_'}),
    ?line '$end_of_table' = ets:next(Tab, First),
    ?line true = ets:info(Tab, fixed),
    ?line ets:fixtable(Tab, false),
    ?line false = ets:info(Tab, fixed),
    ?line ets:delete(Tab).

fixtable_insert(doc) ->    
    ["Check inserts of deleted keys in fixed bag's"];
fixtable_insert(suite) ->
    [];
fixtable_insert(Config) when list(Config) ->
    Ets = make_table(ets, [bag], [{1,1}, {1,2}, {2,1}, {2,2}]),
    ets:fixtable(Ets,true),
    ets:match_delete(Ets,{2,1}),
    2 = ets:first(Ets),
    1 = ets:next(Ets,2),
    ets:delete(Ets,1),
    '$end_of_table' = ets:next(Ets,2),
    ets:insert(Ets, {1,1}),
    1 = ets:next(Ets,2),
    '$end_of_table' = ets:next(Ets,1),
    ets:delete(Ets,1),
    '$end_of_table' = ets:next(Ets,2),
    ets:insert(Ets, {1,2}),
    1 = ets:next(Ets,2),
    '$end_of_table' = ets:next(Ets,1),
    ets:delete(Ets,1),
    '$end_of_table' = ets:next(Ets,2),
    ets:delete(Ets,2),
    '$end_of_table' = ets:next(Ets,2),
    ets:fixtable(Ets,false),
    {'EXIT',{badarg,_}} = (catch ets:next(Ets,2)),
    ok.
    
rename(doc) ->
    ["Check rename of ets tables"];
rename(suite) ->
    [];
rename(Config) when list(Config) ->
    ets:new(foobazz,[named_table, public]),
    ets:insert(foobazz,{foo,bazz}),
    ets:rename(foobazz,ungermanbazz),
    {'EXIT',{badarg, _}} = (catch ets:lookup(foobazz,foo)),
    [{foo,bazz}] = ets:lookup(ungermanbazz,foo),
    ets:delete(ungermanbazz),
    ok.

interface_equality(doc) ->
    ["Tests that the return values and errors are equal for set's and"
     " ordered_set's where applicable"];
interface_equality(suite) ->
    [];
interface_equality(Config) when list(Config) ->
    ?line Set = ets:new(set,[set]),
    ?line OrderedSet = ets:new(ordered_set,[ordered_set]),
    ?line F = fun(X,T,FF) -> case X of 
			   0 -> true; 
			   _ -> 
			       ets:insert(T, {X,
					      integer_to_list(X),
					      X rem 10}),
			       FF(X-1,T,FF) 
		       end 
	end,
    ?line F(100,Set,F),
    ?line F(100,OrderedSet,F),
    ?line equal_results(ets, insert, Set, OrderedSet, [{a,"a"}]),
    ?line equal_results(ets, insert, Set, OrderedSet, [{1,1,"1"}]),
    ?line equal_results(ets, lookup, Set, OrderedSet, [10]),
    ?line equal_results(ets, lookup, Set, OrderedSet, [1000]),
    ?line equal_results(ets, delete, Set, OrderedSet, [10]),
    ?line equal_results(ets, delete, Set, OrderedSet, [nott]),
    ?line equal_results(ets, lookup, Set, OrderedSet, [1000]),
    ?line equal_results(ets, insert, Set, OrderedSet, [10]),
    ?line equal_results(ets, next, Set, OrderedSet, ['$end_of_table']),
    ?line equal_results(ets, prev, Set, OrderedSet, ['$end_of_table']),
    ?line equal_results(ets, match, Set, OrderedSet, [{'_','_','_'}]),
    ?line equal_results(ets, match, Set, OrderedSet, [{'_','_','_','_'}]),
    ?line equal_results(ets, match, Set, OrderedSet, [{$3,$2,2}]),
    ?line equal_results(ets, match, Set, OrderedSet, ['_']),
    ?line equal_results(ets, match, Set, OrderedSet, ['$1']),
    ?line equal_results(ets, match, Set, OrderedSet, [{'_','$50',3}]),
    ?line equal_results(ets, match, Set, OrderedSet, [['_','$50',3]]),
    ?line equal_results(ets, match_delete, Set, OrderedSet, [{'_','_',4}]),
    ?line equal_results(ets, match_delete, Set, OrderedSet, [{'_','_',4}]),
    ?line equal_results(ets, match_object, Set, OrderedSet, [{'_','_',4}]),
    ?line equal_results(ets, match_object, Set, OrderedSet, [{'_','_',5}]),
    ?line equal_results(ets, match_object, Set, OrderedSet, [{'_','_',4}]),
    ?line equal_results(ets, match_object, Set, OrderedSet, ['_']),
    ?line equal_results(ets, match_object, Set, OrderedSet, ['$5011']),
    ?line equal_results(ets, match_delete, Set, OrderedSet, ['$20']),
    ?line equal_results(ets, lookup_element, Set, OrderedSet, [13,2]),
    ?line equal_results(ets, lookup_element, Set, OrderedSet, [13,4]),
    ?line equal_results(ets, lookup_element, Set, OrderedSet, [14,2]),
    ?line equal_results(ets, delete, Set, OrderedSet, []),
    ok.

equal_results(M, F, FirstArg1, FirstArg2 ,ACommon) ->
    Res = maybe_sort((catch apply(M,F, [FirstArg1 | ACommon]))),
    Res = maybe_sort((catch apply(M,F,[FirstArg2 | ACommon]))).

maybe_sort(L) when list(L) ->
    lists:sort(L);
%maybe_sort({'EXIT',{Reason, [{Module, Function, _}|_]}}) ->
%    {'EXIT',{Reason, [{Module, Function, '_'}]}};
maybe_sort({'EXIT',{Reason, List}}) when list(List) ->
    {'EXIT',{Reason, lists:map(fun({Module, Function, _}) ->
				       {Module, Function, '_'}
			       end,
			       List)}};
maybe_sort(Any) ->
    Any.

ordered_match(doc) ->
    ["Test match, match_object and match_delete in ordered set's"];
ordered_match(suite) ->
    [];
ordered_match(Config) when list(Config)->
    ?line F = fun(X,T,FF) -> case X of 
			 0 -> true; 
			 _ -> 
			     ets:insert(T, {X,
					    integer_to_list(X), 
					    X rem 10,
					    X rem 100,
					    X rem 1000}), 
			     FF(X-1,T,FF) 
		     end 
	end,
    ?line T1 = ets:new(xxx,[ordered_set]),
    ?line F(3000,T1,F),
    ?line [[3,3],[3,3],[3,3]] = ets:match(T1, {'_','_','$1','$2',3}),
    ?line F2 = fun(X,Rem,Res,FF) -> case X of 
				  0 -> []; 
				  _ -> 
				      case X rem Rem of
					  Res ->
					      FF(X-1,Rem,Res,FF) ++
						  [{X,
						    integer_to_list(X), 
						    X rem 10,
						    X rem 100,
						    X rem 1000}];
					  _ ->
					      FF(X-1,Rem,Res,FF)
				      end
			      end 
	 end,
    ?line OL1 = F2(3000,100,2,F2),
    ?line OL1 = ets:match_object(T1, {'_','_','_',2,'_'}),
    ?line true = ets:match_delete(T1,{'_','_','_',2,'_'}), 
    ?line [] = ets:match_object(T1, {'_','_','_',2,'_'}),
    ?line OL2 = F2(3000,100,3,F2),
    ?line OL2 = ets:match_object(T1, {'_','_','_',3,'_'}),
    ?line ets:delete(T1),
    ok.
    

ordered(doc) ->
    ["Test basic functionality in ordered_set's."];
ordered(suite) ->
    [];
ordered(Config) when list(Config) ->
    ?line T = ets:new(oset, [ordered_set]),
    ?line InsList = [ 
		      25,26,27,28,
		      5,6,7,8,
		      21,22,23,24,
		      9,10,11,12,
		      1,2,3,4,
		      17,18,19,20,
		      13,14,15,16
		     ],
    ?line lists:foreach(fun(X) ->
			  ets:insert(T,{X,integer_to_list(X)})
		  end,
		  InsList),
    ?line IL2 = lists:map(fun(X) -> {X,integer_to_list(X)} end, InsList),
    ?line L1 = pick_all_forward(T),
    ?line L2 = pick_all_backwards(T),
    ?line S1 = lists:sort(IL2),
    ?line S2 = lists:reverse(lists:sort(IL2)),
    ?line S1 = L1,
    ?line S2 = L2,
    ?line [{1,"1"}] = ets:slot(T,0),
    ?line [{28,"28"}] = ets:slot(T,27),
    ?line 27 = ets:prev(T,28),
    ?line [{7,"7"}] = ets:slot(T,6),
    ?line '$end_of_table' = ets:next(T,28),
    ?line [{12,"12"}] = ets:slot(T,11),
    ?line '$end_of_table' = ets:slot(T,28),
    ?line [{1,"1"}] = ets:slot(T,0),
    ?line 28 = ets:prev(T,29),
    ?line 1 = ets:next(T,0),
    ?line pick_all_forward(T),
    ?line [{7,"7"}] = ets:slot(T,6),
    ?line L2 = pick_all_backwards(T),
    ?line [{7,"7"}] = ets:slot(T,6),
    ok.

pick_all(T,'$end_of_table',How) ->
    [];
pick_all(T,Last,How) ->
    ?line This = case How of
	       next ->
		   ?line ets:next(T,Last);
	       prev ->
		   ?line ets:prev(T,Last)
	   end,
    ?line [LastObj] = ets:lookup(T,Last),
    ?line [LastObj | pick_all(T,This,How)].

pick_all_forward(T) ->
    ?line pick_all(T,ets:first(T),next).
pick_all_backwards(T) ->
    ?line pick_all(T,ets:last(T),prev).
    
    

setbag(doc) ->   ["Small test case for both set and bag type ets tables."];
setbag(suite) -> [];
setbag(Config) when list(Config) ->    
    ?line Set = ets:new(set,[set]),
    ?line Bag = ets:new(bag,[bag]),
    ?line Key = {foo,bar},
    
    %% insert some value
    ?line ets:insert(Set,{Key,val1}),
    ?line ets:insert(Bag,{Key,val1}),
    
    %% insert new value for same key again
    ?line ets:insert(Set,{Key,val2}),
    ?line ets:insert(Bag,{Key,val2}),
    
    %% check
    ?line [{Key,val2}] = ets:lookup(Set,Key),
    ?line case ets:lookup(Bag,Key) of
	[{Key,val1},{Key,val2}] ->
	    ok;
	[{Key,val2},{Key,val1}] ->
	    ok
    end,
    ok.

badnew(doc) ->
    ["Test case to check proper return values for illegal ets:new() calls."];
badnew(suite) -> [];
badnew(Config) when list(Config) ->
    ?line {'EXIT',{badarg,_}} = (catch ets:new(12,[])),
    ?line {'EXIT',{badarg,_}} = (catch ets:new({a,b},[])),
    ?line {'EXIT',{badarg,_}} = (catch ets:new(name,[foo])),
    ok.

verybadnew(doc) ->
    ["Test case to check that a not well formed list does not crash the "
     "emulator. OTP-2314 "];
verybadnew(suite) -> [];
verybadnew(Config) when list(Config) ->
    ?line {'EXIT',{badarg,_}} = (catch ets:new(verybad,[set|protected])),
    ok.

named(doc) ->   ["Small check to see if named tables work."];
named(suite) -> [];
named(Config) when list(Config) ->
    ?line Tab = make_table(foo,
			   [named_table],
			   [{key,val}]),
    ?line [{key,val}] = ets:lookup(foo,key),
    ok.

keypos2(doc) ->   ["Test case to check if specified keypos works."];
keypos2(suite) -> [];
keypos2(Config) when list(Config) ->
    ?line Tab = make_table(foo,
			   [set,{keypos,2}],
			   [{val,key}, {val2,key}]),
    ?line [{val2,key}] = ets:lookup(Tab,key),
    ok.

privacy(doc) ->
    ["Privacy check. Check that a named(public/private/protected) table "
     "cannot be read by",
     "the wrong process(es)."];
privacy(suite) -> [];
privacy(Config) when list(Config) ->
    ?line process_flag(trap_exit,true),
    ?line Nosy = spawn_link(?MODULE,nosy,[self()]),
    receive
	{'EXIT',Nosy,Reason} ->
	    ?line exit({privacy_test,Reason});
	ok ->
	    ok
    end,

    %% check read rights
    ?line [] = ets:lookup(pub, foo),
    ?line [] = ets:lookup(prot,foo),
    ?line {'EXIT',{badarg,_}} = (catch ets:lookup(priv,foo)),

    %% check write rights
    ?line true = ets:insert(pub, {1,foo}),
    ?line {'EXIT',{badarg,_}} = (catch ets:insert(prot,{2,foo})),
    ?line {'EXIT',{badarg,_}} = (catch ets:insert(priv,{3,foo})),

    %% check that it really wasn't written, either
    ?line [] = ets:lookup(prot,foo),

    Nosy ! die,
    receive
	{'EXIT',Nosy,_} ->
	    ok
    end,
    ok.

nosy(Boss) ->
    Pub  = ets:new(pub, [public,named_table]),
    Prot = ets:new(prot,[protected,named_table]),
    Priv = ets:new(priv,[private,named_table]),
    Boss ! ok,
    receive
	die ->
	    ok
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

insert(doc) ->   ["Test proper and improper inserts into a table."];
insert(suite) -> [empty,badinsert].

empty(doc) ->
    ["Check lookup in an empty table and lookup of a non-existing key"];
empty(suite) -> [];
empty(Config) when list(Config) ->
    ?line Tab = ets:new(foo,[]),
    ?line [] = ets:lookup(Tab,key),
    ?line true = ets:insert(Tab,{key2,val}),
    ?line [] = ets:lookup(Tab,key),
    ok.

badinsert(doc) ->
    ["Check proper return values for illegal insert operations."];
badinsert(suite) -> [];
badinsert(Config) when list(Config) ->
    ?line {'EXIT',{badarg,_}} = (catch ets:insert(foo,{key,val})),
    
    ?line Tab = ets:new(foo,[]),
    ?line {'EXIT',{badarg,_}} = (catch ets:insert(Tab,{})),

    ?line Tab3 = ets:new(foo,[{keypos,3}]),
    ?line {'EXIT',{badarg,_}} = (catch ets:insert(Tab3,{a,b})),

    ?line {'EXIT',{badarg,_}} = (catch ets:insert(Tab,[key,val2])),
    ok.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

lookup(doc) -> ["Some tests for lookups (timing, bad lookups, etc.)."];
lookup(suite) -> [time_lookup,badlookup].

time_lookup(doc) ->   ["Lookup timing."];
time_lookup(suite) -> [];
time_lookup(Config) when list(Config) ->
    %% just for timing, really
    ?line Tab = ets:new(foo,[]),
    ?line fill_tab(Tab,foo),
    ?line ets:insert(Tab,{{a,key},foo}),
    ?line {Time,_} = ?t:timecall(test_server,do_times,
				    [10000,ets,lookup,[Tab,{a,key}]]),
    ?line {comment,lists:flatten(io_lib:format(
				   "~.1f ets lookups/s",[10000/Time]))}.

badlookup(doc) ->
    ["Check proper return values from bad lookups in existing/non existing "
     " ets tables"];
badlookup(suite) -> [];
badlookup(Config) when list(Config) ->
    ?line {'EXIT',{badarg,_}} = (catch ets:lookup(foo,key)),
    
    ?line Tab = ets:new(foo,[]),
    ?line ets:delete(Tab),
    ?line {'EXIT',{badarg,_}} = (catch ets:lookup(Tab,key)),
    ok.

fill_tab(Tab,Val) ->
    ?line ets:insert(Tab,{key,Val}),
    ?line ets:insert(Tab,{{a,144},Val}),
    ?line ets:insert(Tab,{{a,key2},Val}),
    ?line ets:insert(Tab,{14,Val}).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

lookup_element(doc) -> ["Some tests for lookup_element."];
lookup_element(suite) -> [lookup_element_mult].

lookup_element_mult(doc) ->   ["Multiple return elements (OTP-2386)"];
lookup_element_mult(suite) -> [];
lookup_element_mult(Config) when list(Config) ->
    ?line T = ets:new(service, [bag, {keypos, 2}]),
    ?line D = lists:reverse(lem_data()),
    ?line lists:foreach(fun(X) -> ets:insert(T, X) end, D),
    ?line ok = lem_crash_3(T).


lem_data() ->
    [
     {service,'eddie2@boromir',{150,236,14,103},httpd88,self()},
     {service,'eddie2@boromir',{150,236,14,103},httpd80,self()},
     {service,'eddie3@boromir',{150,236,14,107},httpd88,self()},
     {service,'eddie3@boromir',{150,236,14,107},httpd80,self()},
     {service,'eddie4@boromir',{150,236,14,108},httpd88,self()}
    ].

lem_crash(T) ->
    L = ets:lookup_element(T, 'eddie2@boromir', 3),
    {erlang:hash(L, 256), L}.

lem_crash_3(T) ->
    lem_crash(T),
    io:format("Survived once~n"),
    lem_crash(T),
    io:format("Survived twice~n"),
    lem_crash(T),
    io:format("Survived all!~n"),
    ok.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

delete(doc) -> ["Check delete functionality (proper/improper deletes)"];
delete(suite) -> [delete_elem,delete_tab,baddelete,match_delete,match_delete3].

delete_elem(doc) ->
    ["Check delete of an element inserted in a `filled' table."];
delete_elem(suite) -> [];
delete_elem(Config) when list(Config) ->
    Types = [set, bag, duplicate_bag, ordered_set],
    delete_elem2(Types).
delete_elem2([]) ->
    ok;
delete_elem2([Type|Tail]) ->
    ?line Tab = ets:new(foo,[Type]),
    ?line fill_tab(Tab,foo),
    ?line ets:insert(Tab,{{b,key},foo}),
    ?line ets:insert(Tab,{{c,key},foo}),
    ?line true = ets:delete(Tab,{b,key}),
    ?line [] = ets:lookup(Tab,{b,key}),
    ?line [{{c,key},foo}] = ets:lookup(Tab,{c,key}),
    delete_elem2(Tail).
    
delete_tab(doc) ->
    ["Check that ets:delete() works and releases the name of the deleted "
     "table."];
delete_tab(suite) -> [];
delete_tab(Config) when list(Config) ->
    ?line Nam1 = ets:new(foo,[named_table]),
    ?line true = ets:delete(foo),
    %% name should be available again, now
    ?line Nam2 = ets:new(foo,[named_table]),
    ?line true = ets:delete(Nam2),
    ok.

baddelete(doc) ->
    ["Check proper return values for illegal delete operations."];
baddelete(suite) -> [];
baddelete(Config) when list(Config) ->
    ?line {'EXIT',{badarg,_}} = (catch ets:delete(foo)),
    ?line Tab = ets:new(foo,[]),
    ?line true = ets:delete(Tab),
    ?line {'EXIT',{badarg,_}} = (catch ets:delete(Tab)),
    ok.

match_delete(doc) ->
    ["Check that match_delete works. Also tests tab2list function."];
match_delete(suite) -> [];    
match_delete(Config) when list(Config) ->
    Types = [set, bag, duplicate_bag, ordered_set],
    match_delete2(Types).

match_delete2([]) -> ok;
match_delete2([Type|Tail]) ->
    ?line Tab = ets:new(kad,[Type]),
    ?line fill_tab(Tab,foo),
    ?line ets:insert(Tab,{{c,key},bar}),
    ?line _ = ets:match_delete(Tab,{'_',foo}),
    ?line [{{c,key},bar}] = ets:tab2list(Tab),
    ?line _ = ets:match_delete(Tab,'_'),
    ?line [] = ets:tab2list(Tab),
    ?line true = ets:delete(Tab),
    match_delete2(Tail).

match_delete3(doc) ->
    ["OTP-3005: check match_delete with constant argument."];
match_delete3(suite) -> [];    
match_delete3(Config) when list(Config) ->
    T = make_table(test,
		   [duplicate_bag],
		   [{aa,17},
		    {cA,1000},
		    {cA,17},
		    {cA,1000},
		    {aa,17}]),
    %% 'aa' and 'cA' have the same hash value in the current
    %% implementation. This causes the aa's to precede the cA's, to make
    %% the test more interesting.
    [{cA,1000},{cA,1000}] = ets:match_object(T, {'_', 1000}),
    ets:match_delete(T, {cA,1000}),
    [] = ets:match_object(T, {'_', 1000}),
    ets:delete(T),
    true.


%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

firstnext(doc) -> ["Tests ets:first/1 & ets:next/2."];
firstnext(suite) -> [];
firstnext(Config) when list(Config) ->
    ?line Tab = ets:new(foo,[]),
    ?line [] = firstnext_collect(Tab,ets:first(Tab),[]),
    ?line fill_tab(Tab,foo),
    ?line Len = length(ets:tab2list(Tab)),
    ?line Len = length(firstnext_collect(Tab,ets:first(Tab),[])),
    ok.

firstnext_collect(Tab,'$end_of_table',List) ->
    ?line List;
firstnext_collect(Tab,Key,List) ->
    ?line firstnext_collect(Tab,ets:next(Tab,Key),[Key|List]).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

all_2214(doc) ->
    ["all/1's result should not contain 'undefined' (OTP-2214)"];
all_2214(suite) ->
    [];
all_2214(Config) when list(Config) ->
    ?line Pid = spawn(?MODULE, do0_2214, []),
    ?line ok = check_2214(1),
    ?line exit(Pid, done),			%kill off the ets tables
    ok.

do0_2214() ->
    spawn_link(?MODULE, do1_2214, [self()]).

do_2214(Pid) ->
    spawn_link(?MODULE, do1_2214, [Pid]).

do1_2214(Pid) ->
    ets:new(fnaff, [bag]),
    receive
    after 2000 -> ok
    end,
    do_2214(Pid).

check_2214(I) when I > 100000 ->
    ok;
check_2214(I) ->
    A = ets:all(),
    case undef_check(A, A) of
	true ->
	    {fail, I, A};
	false ->
	    check_2214(I+1)
    end.

undef_check([], L) ->
    false;
undef_check([{T, undefined} | _], L) ->
    true;
undef_check([_ | Rest], L) ->
    undef_check(Rest, L).

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

slot(suite) -> [];
slot(Config) when list(Config) ->
    ?line Tab = ets:new(foo,[]),
    ?line fill_tab(Tab,foo),
    ?line Elts = ets:info(Tab,size),
    ?line Elts = slot_loop(Tab,0,0),
    ok.

slot_loop(Tab,SlotNo,EltsSoFar) ->
    ?line case ets:slot(Tab,SlotNo) of
	      '$end_of_table' ->
		  ?line {'EXIT',{badarg,_}} =
		      (catch ets:slot(Tab,SlotNo+1)),
		  ?line EltsSoFar;
	      Elts ->
		  ?line slot_loop(Tab,SlotNo+1,EltsSoFar+length(Elts))
    end.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

match(suite) -> [match1, match2, match_object, match_object2].

match1(suite) -> [];
match1(Config) when list(Config) ->
    ?line Tab = ets:new(foo,[]),
    ?line fill_tab(Tab,foo),
    ?line [] = ets:match(Tab,{}),
    ?line ets:insert(Tab,{{one,4},4}),
    ?line ets:insert(Tab,{{one,5},5}),
    ?line ets:insert(Tab,{{two,4},4}),
    ?line ets:insert(Tab,{{two,5},6}),
    ?line case ets:match(Tab,{{one,'_'},'$0'}) of
	      [[4],[5]] -> ok;
	      [[5],[4]] -> ok
	  end,
    ?line case ets:match(Tab,{{two,'$1'},'$0'}) of
	      [[4,4],[6,5]] -> ok;
	      [[6,5],[4,4]] -> ok
	  end,
    ?line case ets:match(Tab,{{two,'$9'},'$4'}) of
	      [[4,4],[6,5]] -> ok;
	      [[6,5],[4,4]] -> ok
	  end,
    ?line case ets:match(Tab,{{two,'$9'},'$22'}) of
	      [[4,4],[5,6]] -> ok;
	      [[5,6],[4,4]] -> ok
	  end,
    ?line [[4]] = ets:match(Tab,{{two,'$0'},'$0'}),
    ?line Len = length(ets:match(Tab,'$0')),
    ?line Len = length(ets:match(Tab,'_')),
    ?line if Len > 4 -> ok end,
    ?line true = ets:delete(Tab),
    ok.

match2(doc) -> ["Tests match with specified keypos bag table."];
match2(suite) -> [];
match2(Config) when list(Config) ->
    ?line Tab = make_table(foobar,
			   [bag, private, named_table, {keypos, 2}],
			   [{value1, key1},
			    {value2_1, key2},
			    {value2_2, key2},
			    {value3_1, key3},
			    {value3_2, key3}]),
    ?line case length(ets:match(Tab, '$1')) of
	      5 -> ok;
	      _ -> ?t:fail("Length of matched list is wrong.")
	  end,
    ?line [[value3_1],[value3_2]] = ets:match(Tab, {'$1', key3}),
    ?line [[key1]] = ets:match(Tab, {value1, '$1'}),
    ?line true = ets:delete(Tab),
    ok.

match_object(doc) -> ["Sime ets:match_object test."];
match_object(suite) -> [];    
match_object(Config) when list(Config) ->
    ?line Tab = ets:new(foobar, []),
    ?line fill_tab(Tab, foo),
    ?line ets:insert(Tab, {{one, 4}, 4}),
    ?line ets:insert(Tab,{{one,5},5}),
    ?line ets:insert(Tab,{{two,4},4}),
    ?line ets:insert(Tab,{{two,5},6}),
    ?line case ets:match_object(Tab, {{one, '_'}, '$0'}) of
	[{{one,5},5},{{one,4},4}] -> ok;
	[{{one,4},4},{{one,5},5}] -> ok;
	_ -> ?t:fail("ets:match_object() returned something funny.")
    end,
    ?line case ets:match_object(Tab, {{two, '$1'}, '$0'}) of
	[{{two,5},6},{{two,4},4}] -> ok;
	[{{two,4},4},{{two,5},6}] -> ok;
	_ -> ?t:fail("ets:match_object() returned something funny.")
    end,
    ?line case ets:match_object(Tab, {{two, '$9'}, '$4'}) of
	[{{two,5},6},{{two,4},4}] -> ok;
	[{{two,4},4},{{two,5},6}] -> ok;
	_ -> ?t:fail("ets:match_object() returned something funny.")
    end,
    ?line case ets:match_object(Tab, {{two, '$9'}, '$22'}) of
	[{{two,5},6},{{two,4},4}] -> ok;
	[{{two,4},4},{{two,5},6}] -> ok;
	_ -> ?t:fail("ets:match_object() returned something funny.")
    end,
    % Check that unsucessful match returns an empty list.
    ?line [] = ets:match_object(Tab, {{three,'$0'}, '$92'}),
    % Check that '$0' equals '_'.
    Len = length(ets:match_object(Tab, '$0')),
    Len = length(ets:match_object(Tab, '_')),
    ?line if Len > 4 -> ok end,
    ?line true = ets:delete(Tab),
    ok.

match_object2(suite) -> [];
match_object2(doc) -> ["Tests that db_match_object does not generate "
		       "a `badarg' when resuming a search with no "
		       "previous matches."];
match_object2(Config) when list(Config) ->
    ?line Tab = ets:new(foo, [bag, {keypos, 2}]),
    ?line fill_tab2(Tab, 0, 13005),     % match_db_object does 1000
                                       % elements per pass, might
                                       % change in the future.
    ?line case catch ets:match_object(Tab, {hej, '$1'}) of
	      {'EXIT', _} ->
		  ets:delete(Tab),
		  ?t:fail("match_object EXIT:ed");
	      [] ->
		  io:format("Nothing matched.");
	      List ->
		  io:format("Matched:~p~n",[List])
	  end,
    ets:delete(Tab),
    ok.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

misc(suite) -> [misc1, fixtable, info, dups, tab2list].

tab2list(doc) -> ["Tests tab2list (OTP-3319)"];
tab2list(suite) -> [];
tab2list(Config) when list(Config) ->
    ?line Tab = make_table(foo,
			   [ordered_set],
			   [{a,b}, {c,b}, {b,b}, {a,c}]),
    ?line [{a,c},{b,b},{c,b}] = ets:tab2list(Tab),
    ok.

misc1(doc) -> ["Simple general small test. ",
	      "If this fails, ets is in really bad shape."];
misc1(suite) -> [];
misc1(Config) when list(Config) ->
    ?line Tab = ets:new(foo,[]),
    ?line true = lists:member(Tab,ets:all()),
    ?line ets:delete(Tab),
    ?line false = lists:member(Tab,ets:all()),
    ?line case catch ets:delete(Tab) of
	      {'EXIT', Reason} ->
		  ok;
	      true ->
		  ?t:fail("Delete of nonexisting table returned `true'.")
	  end,
    ok.

fixtable(doc) ->  ["Check the fixtable function."];
fixtable(suite) -> [];
fixtable(Config) when list(Config) ->
    ?line Tab = ets:new(foo, []),
    ?line fill_tab(Tab, foobar),
    ?line true = ets:fixtable(Tab, true),
    ?line receive after 1 -> ok end,
    ?line true = ets:fixtable(Tab, false),
    ?line ets:delete(Tab),
    ?line case catch ets:fixtable(Tab, true) of
	      {'EXIT', Reason} ->
		  ok;
	      _ ->
		  ?t:fail("Fixtable on nonexisting table returned `true'")
	  end,
    ok.

info(doc) -> ["Tests ets:info result for required tuples."];
info(suite) -> [];
info(Config) when list(Config) ->
    ?line MeMyselfI=self(),
    ?line ThisNode=node(),
    ?line Tab = ets:new(foobar, [{keypos, 2}]),
    ?line Res = tuple_to_list(ets:info(Tab)),
    ?line {value, {memory, Mem}} = lists:keysearch(memory, 1, Res),
    ?line {value, {owner, MeMyselfI}} = lists:keysearch(owner, 1, Res),
    ?line {value, {name, foobar}} = lists:keysearch(name, 1, Res),
    ?line {value, {size, 0}} = lists:keysearch(size, 1, Res),
    ?line {value, {node, ThisNode}} = lists:keysearch(node, 1, Res),
    ?line {value, {named_table, false}} = lists:keysearch(named_table, 1, Res),
    ?line {value, {type, set}} = lists:keysearch(type, 1, Res),
    ?line {value, {keypos, 2}} = lists:keysearch(keypos, 1, Res),
    ?line {value, {protection, protected}} =
	lists:keysearch(protection, 1, Res),
    ?line true = ets:delete(Tab),
    ?line undefined = ets:info(non_existing_table_xxyy),
    ?line undefined = ets:info(non_existing_table_xxyy,type),
    ?line undefined = ets:info(non_existing_table_xxyy,node),
    ?line undefined = ets:info(non_existing_table_xxyy,named_table),
    ?line undefined = ets:info(non_existing_table_xxyy,safe_fixed),
    ok.

dups(doc) -> ["Test various duplicate_bags stuff"];
dups(suite) -> [];
dups(Config) when list(Config) ->
    ?line T = make_table(funky,
			 [duplicate_bag],
			 [{1, 2}, {1, 2}]),
    ?line 2 = length(ets:tab2list(T)),
    ?line ets:delete(T, 1),
    ?line [] = ets:lookup(T, 1),

    ?line ets:insert(T, {1, 2, 2}),
    ?line ets:insert(T, {1, 2, 4}),
    ?line ets:insert(T, {1, 2, 2}),
    ?line ets:insert(T, {1, 2, 2}),
    ?line ets:insert(T, {1, 2, 4}),

    ?line 5 = length(ets:tab2list(T)),

    ?line 5 = length(ets:match(T, {'$1', 2, '$2'})),
    ?line 3 = length(ets:match(T, {'_', '$1', '$1'})),
    ?line ets:match_delete(T, {'_', '$1', '$1'}),
    ?line 0 = length(ets:match(T, {'_', '$1', '$1'})),
    ?line ets:delete(T),
    ok.

%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

files(suite) -> [tab2file, tab2file2, tab2file3].

tab2file(doc) -> ["Check the ets:tab2file function on an empty "
		  "ets table."];
tab2file(suite) -> [];
tab2file(Config) when list(Config) ->
    %% Write an empty ets table to a file, read back and check properties.
    ?line Tab = ets:new(ets_SUITE_foo_tab, [named_table, set, private,
					    {keypos, 2}]),
    ?line FName = filename:join([?config(priv_dir, Config),"tab2file_case"]),
    ?line ok = ets:tab2file(Tab, FName),
    ?line true = ets:delete(Tab),
    %
    ?line {ok, Tab2} = ets:file2tab(FName),
    ?line private = ets:info(Tab2, protection),
    ?line true = ets:info(Tab2, named_table),
    ?line 2 = ets:info(Tab2, keypos),
    ?line set = ets:info(Tab2, type),
    ?line true = ets:delete(Tab2),
    ok.
    
tab2file2(doc) -> ["Check the ets:tab2file function on a ",
		   "filled set type ets table."];
tab2file2(suite) -> [];
tab2file2(Config) when list(Config) ->	
    %% Try the same on a filled set table.
    ?line Tab = ets:new(ets_SUITE_foo_tab, [named_table, set, private,
					    {keypos, 2}]),
    ?line FName = filename:join([?config(priv_dir, Config),"tab2file2_case"]),
    ?line ok = fill_tab2(Tab, 0, 10000),   % Fill up the table (grucho mucho!)
    ?line Len = length(ets:tab2list(Tab)),
    ?line ok = ets:tab2file(Tab, FName),
    ?line true = ets:delete(Tab),
    %
    ?line {ok, Tab2} = ets:file2tab(FName),
    ?line private = ets:info(Tab2, protection),
    ?line true = ets:info(Tab2, named_table),
    ?line 2 = ets:info(Tab2, keypos),
    ?line set = ets:info(Tab2, type),
    ?line Len = length(ets:tab2list(Tab2)),
    ?line true = ets:delete(Tab2),
    ok.

tab2file3(doc) -> ["Check the ets:tab2file function on a ",
		   "filled bag type ets table."];
tab2file3(suite) -> [];
tab2file3(Config) when list(Config) ->
    %% Try the same on a filled bag table.
    ?line Tab = ets:new(ets_SUITE_foo_tab, [named_table, bag, private,
					    {keypos, 2}]),
    ?line FName = filename:join([?config(priv_dir, Config),"tab2file3_case"]),
    ?line ok = fill_tab2(Tab, 0, 10000),   % Fill up the table (grucho mucho!)
    ?line Len = length(ets:tab2list(Tab)),
    ?line Mem = ets:info(Tab, memory),
    ?line ok = ets:tab2file(Tab, FName),
    ?line true = ets:delete(Tab),

    ?line {ok, Tab2} = ets:file2tab(FName),
    ?line private = ets:info(Tab2, protection),
    ?line true = ets:info(Tab2, named_table),
    ?line 2 = ets:info(Tab2, keypos),
    ?line bag = ets:info(Tab2, type),
    ?line Len = length(ets:tab2list(Tab2)),
    ?line Mem = ets:info(Tab2, memory),
    ?line true = ets:delete(Tab2),
    ok.

-define(test_list, [8,5,4,1,58,125,255, 250, 245, 240, 235,
		    230, Num rem 255, 255, 125, 130, 135, 140, 145,
		    150, 134, 12, 54, Val rem 255, 12, 3, 6, 9, 126]).
-define(big_test_list, [Num rem 256|lists:seq(1, 66)]).
-define(test_integer, 2846287468+Num).
-define(test_float, 187263.18236-Val).
-define(test_atom, some_crazy_atom).
-define(test_tuple, {just, 'Some', 'Tuple', 1, [list, item], Val+Num}).

%% Insert different datatypes into a ets table.
fill_tab2(Tab, Val, 0) ->
    ok;
fill_tab2(Tab, Val, Num) ->
    ?line Item =
	case Num rem 10 of
	    0 -> "String";
	    1 -> ?line ?test_atom;
	    2 -> ?line ?test_tuple;
	    3 -> ?line ?test_integer;
	    4 -> ?line ?test_float;
	    5 -> ?line list_to_binary(?test_list); %Heap binary
	    6 -> ?line list_to_binary(?big_test_list); %Refc binary
	    7 -> ?line make_sub_binary(?test_list, Num); %Sub binary
	    8 -> ?line ?test_list;
	    9 -> ?line fun(X) -> {Tab,Val,X*Num} end
	end,
    ?line true=ets:insert(Tab, {Item, Val}),
    ?line fill_tab2(Tab, Val+1, Num-1),
    ok.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

make_sub_binary(List, Num) when list(List) ->
    N = Num rem 23,
    Bin = list_to_binary([lists:seq(0, N)|List]),
    {_,B} = split_binary(Bin, N+1),
    B.

heavy(suite) -> [heavy_lookup, heavy_lookup_element].

%% Lookup stuff like crazy...
heavy_lookup(doc) -> ["Performs multiple lookups for every key ",
		      "in a large table."];
heavy_lookup(suite) -> [];
heavy_lookup(Config) when list(Config) ->
    ?line Tab = ets:new(foobar_table, [set, protected, {keypos, 2}]),
    ?line ok = fill_tab2(Tab, 0, 7000),
    ?line ?t:do_times(50, ?MODULE, do_lookup, [Tab, 6999]),
    ?line true = ets:delete(Tab),
    ok.

do_lookup(Tab, 0) -> ok;
do_lookup(Tab, N) ->
    case ets:lookup(Tab, N) of
	[] -> ?t:format("Set #~p was reported as empty. Not valid.",
				 [N]),
	      exit('Invalid lookup');
	_ ->  do_lookup(Tab, N-1)
    end.

heavy_lookup_element(doc) -> ["Performs multiple lookups for ",
			      "every element in a large table."];
heavy_lookup_element(suite) -> [];
heavy_lookup_element(Config) when list(Config) ->
    ?line Tab = ets:new(foobar_table, [set, protected, {keypos, 2}]),
    ?line ok = fill_tab2(Tab, 0, 7000),
    case os:type() of
	vxworks ->
	    ?line ?t:do_times(5, ?MODULE, do_lookup_element,
				       [Tab, 6999, 1]);
						% lookup ALL elements 5 times.
	_ ->
	    ?line ?t:do_times(50, ?MODULE, do_lookup_element,
				       [Tab, 6999, 1])
						% lookup ALL elements 50 times.
    end,
    ?line true = ets:delete(Tab),
    ok.

do_lookup_element(Tab, 0, _) -> ok;
do_lookup_element(Tab, N, M) ->
    ?line case catch ets:lookup_element(Tab, N, M) of
	      {'EXIT', {badarg, _}} ->
		  case M of
		      1 -> ?t:fail("Set #~p reported as empty. Not valid.",
				   [N]),
			   exit('Invalid lookup_element');
		      _ -> ?line do_lookup_element(Tab, N-1, 1)
		  end;
	      _ -> ?line do_lookup_element(Tab, N, M+1)
    end.


fold(suite) -> [foldl_ordered, foldr_ordered,
		foldl, foldr,
		fold_empty].

fold_empty(doc) ->
    [];
fold_empty(suite) -> [];
fold_empty(Config) when list(Config) ->
    ?line Tab = make_table(a, [], []),
    ?line [] = ets:foldl(fun(X) -> exit(hej) end, [], Tab),
    ?line [] = ets:foldr(fun(X) -> exit(hej) end, [], Tab),
    ok.

foldl(doc) ->
    [];
foldl(suite) -> [];
foldl(Config) when list(Config) ->
    ?line L = [{a,1}, {c,3}, {b,2}],
    ?line LS = lists:sort(L),
    ?line Tab = make_table(a, [bag], L),
    ?line LS = lists:sort(ets:foldl(fun(E,A) -> [E|A] end, [], Tab)),
    ok.

foldr(doc) ->
    [];
foldr(suite) -> [];
foldr(Config) when list(Config) ->
    ?line L = [{a,1}, {c,3}, {b,2}],
    ?line LS = lists:sort(L),
    ?line Tab = make_table(a, [bag], L),
    ?line LS = lists:sort(ets:foldr(fun(E,A) -> [E|A] end, [], Tab)),
    ok.

foldl_ordered(doc) ->
    [];
foldl_ordered(suite) -> [];
foldl_ordered(Config) when list(Config) ->
    ?line L = [{a,1}, {c,3}, {b,2}],
    ?line LS = lists:sort(L),
    ?line Tab = make_table(a, [ordered_set], L),
    ?line LS = lists:reverse(ets:foldl(fun(E,A) -> [E|A] end, [], Tab)),
    ok.

foldr_ordered(doc) ->
    [];
foldr_ordered(suite) -> [];
foldr_ordered(Config) when list(Config) ->
    ?line L = [{a,1}, {c,3}, {b,2}],
    ?line LS = lists:sort(L),
    ?line Tab = make_table(a, [ordered_set], L),
    ?line LS = ets:foldr(fun(E,A) -> [E|A] end, [], Tab),
    ok.

% i_2409(doc) ->
%     ["Test that ets:i() does not starve out other processes which",
%      "might cause heartbeat timeout",
%      "OTP-2409"];
% i_2409(suite) -> [];
% i_2409(Config) when list(Config) ->
%     ?line Tab = ets:new(heartbeat_table, [set, protected, {keypos, 2}]),
%     ?line ok = fill_tab2(Tab, 0, 20000),
%     Pid = spawn(?MODULE,heartbeat_loop,[0]),
%     ?line true = eloop(Tab,50),
%     Pid ! {self(),stop},
%     Count = receive 
% 	{Pid, _Count} ->
% 	    _Count
%     end,
%     ?line C = if 
% 	      Count > 0 ->
% 		  ok;
% 	      true ->
% 		  false
% 	  end,
%     ?line io:format("Count = ~p~n",[Count]),
%     ?line ok = C,
%     ets:delete(Tab).


member(suite) ->
    [];
member(doc) ->
    ["Tests ets:member BIF"];
member(Config) when list(Config) ->
    ?line T = ets:new(xxx,[ordered_set]),
    ?line U = ets:new(xxx,[]),
    ?line false = ets:member(T,hej),
    ?line false = ets:member(U,hej),
    ?line E = fun(0,F)->ok;
		 (N,F) -> 
		      ?line ets:insert(T,{N,N rem 10}), 
		      ?line ets:insert(U,{N,N rem 10}), 
		      F(N-1,F) 
	      end,
    ?line E(10000,E),                                                          
    ?line false = ets:member(T,hej),
    ?line false = ets:member(U,hej),
    ?line true = ets:member(T,1),
    ?line true = ets:member(U,1),
    ?line false = ets:member(T,20000),
    ?line false = ets:member(U,20000),
    ?line ets:delete(T,5),
    ?line ets:delete(U,5),
    ?line false = ets:member(T,5),
    ?line false = ets:member(U,5),
    ?line ets:safe_fixtable(T,true),
    ?line ets:safe_fixtable(U,true),
    ?line ets:delete(T,6),
    ?line ets:delete(U,6),
    ?line false = ets:member(T,6),
    ?line false = ets:member(U,6),
    ?line ets:safe_fixtable(T,false),
    ?line ets:safe_fixtable(U,false),
    ?line false = ets:member(T,6),
    ?line false = ets:member(U,6),
    ?line ets:delete(T),
    ?line ets:delete(U),
    ?line {'EXIT',{badarg,_}} = (catch ets:member(finnsinte, 23)),
    ?line {'EXIT',{badarg,_}} = (catch ets:member(T, 23)),
    ?line {'EXIT',{badarg,_}} = (catch ets:member(U, 23)),
    ok.


% eloop(Tab,0) ->
%     true;
% eloop(Tab,Count) ->
%     ets:info(Tab,memory),
%     eloop(Tab,Count-1).

% heartbeat_loop(Count) -> 
%     receive 
% 	{From,stop} ->
% 	    From ! {self(),Count}
%     after 1 ->
% 	    heartbeat_loop(Count+1)
%     end.

       
build_table(L1,L2,Num) ->
    T = ets:new(xxx, [ordered_set]
	       ),
    lists:foreach(
      fun(X1) ->
	      lists:foreach(
		fun(X2) ->
			F = fun(FF,N) ->
				    ets:insert(T,{{X1,X2,N}, 
						  X1, X2, N}),
				    case N of 
					0 ->
					    ok;
					_ ->
					    FF(FF,N-1)
				    end
			    end,
			F(F,Num)
		end, L2)
      end, L1),
    T.

build_table2(L1,L2,Num) ->
    T = ets:new(xxx, [ordered_set]
	       ),
    lists:foreach(
      fun(X1) ->
	      lists:foreach(
		fun(X2) ->
			F = fun(FF,N) ->
				    ets:insert(T,{{N,X1,X2}, 
						  N, X1, X2}),
				    case N of 
					0 ->
					    ok;
					_ ->
					    FF(FF,N-1)
				    end
			    end,
			F(F,Num)
		end, L2)
      end, L1),
    T.

time_match_object(Tab,Match, Res) ->
    T1 = erlang:now(),
    Res = ets:match_object(Tab,Match),
    T2 = erlang:now(),
    nowdiff(T1,T2).

time_match(Tab,Match) ->
    T1 = erlang:now(),
    ets:match(Tab,Match),
    T2 = erlang:now(),
    nowdiff(T1,T2).

seventyfive_percent_success(_,S,Fa,0) ->
    true = (S > ((S + Fa) * 0.75));

seventyfive_percent_success({M,F,A},S,Fa,N) ->
    case (catch apply(M,F,A)) of
	{'EXIT', _} ->
	    seventyfive_percent_success({M,F,A},S,Fa+1,N-1);
	_ ->
	    seventyfive_percent_success({M,F,A},S+1,Fa,N-1)
    end.


nowtonumber({Mega, Secs, Milli}) ->
    Milli + Secs * 1000000 + Mega * 1000000000000.
nowdiff(T1,T2) ->
    nowtonumber(T2) - nowtonumber(T1).

create_random_string(0) ->
    [];

create_random_string(OfLength) ->
    C = case random:uniform(2) of
	1 ->
	    (random:uniform($Z - $A + 1) - 1) + $A;
	_ ->
	    (random:uniform($z - $a + 1) - 1) + $a
	end,
    [C | create_random_string(OfLength - 1)].


create_random_tuple(OfLength) ->
    list_to_tuple(lists:map(fun(X) ->
				    list_to_atom([X])
			    end,create_random_string(OfLength))).

create_partly_bound_tuple(OfLength) ->
    case random:uniform(2) of
	1 ->
	   create_partly_bound_tuple1(OfLength); 
	_ ->
	   create_partly_bound_tuple3(OfLength)
    end.

create_partly_bound_tuple1(OfLength) ->
    T0 = create_random_tuple(OfLength),
    I = random:uniform(OfLength),
    setelement(I,T0,'$1').


set_n_random_elements(T0,0,_,_) ->
    T0;
set_n_random_elements(T0,N,OfLength,GenFun) ->
    I = random:uniform(OfLength),
    What = GenFun(I),
    case element(I,T0) of
	What ->
	    set_n_random_elements(T0,N,OfLength,GenFun);
	Else ->
	    set_n_random_elements(setelement(I,T0,What),
				  N-1,OfLength,GenFun)
    end.

make_dollar_atom(I) ->
    list_to_atom([$$] ++ integer_to_list(I)).
create_partly_bound_tuple2(OfLength) ->
    T0 = create_random_tuple(OfLength),
    I = random:uniform(OfLength - 1),
    set_n_random_elements(T0,I,OfLength,fun make_dollar_atom/1).

create_partly_bound_tuple3(OfLength) ->
    T0 = create_random_tuple(OfLength),
    I = random:uniform(OfLength - 1),
    set_n_random_elements(T0,I,OfLength,fun(_) -> '_' end).

do_n_times(_,0) ->
    ok;
do_n_times(Fun,N) ->
    Fun(),
    case N rem 1000 of
	0 ->
	    io:format(".");
	_ ->
	    ok
    end,
    do_n_times(Fun,N-1).

make_table(Name, Options, Elements) ->
    T = ets:new(Name, Options),
    lists:foreach(fun(E) -> ets:insert(T, E) end, Elements),
    T.
filltabint(Tab,0) ->
    Tab;
filltabint(Tab,N) ->
    ets:insert(Tab,{N,integer_to_list(N)}),
    filltabint(Tab,N-1).
filltabint2(Tab,0) ->
    Tab;
filltabint2(Tab,N) ->
    ets:insert(Tab,{N + N rem 2,integer_to_list(N)}),
    filltabint2(Tab,N-1).
filltabint3(Tab,0) ->
    Tab;
filltabint3(Tab,N) ->
    ets:insert(Tab,{N + N rem 2,integer_to_list(N + N rem 2)}),
    filltabint3(Tab,N-1).
xfilltabint(Tab,N) ->
    case ets:info(Tab,type) of
	bag ->
	    filltabint2(Tab,N);
	duplicate_bag ->
	    ets:select_delete(Tab,[{'_',[],[true]}]),
	    filltabint3(Tab,N);
	_ ->
	    filltabint(Tab,N)
    end.
    
filltabstr(Tab,0) ->
    Tab;
filltabstr(Tab,N) ->
    ets:insert(Tab,{integer_to_list(N),N}),
    filltabstr(Tab,N-1).
filltabstr2(Tab,0) ->
    Tab;
filltabstr2(Tab,N) ->
    ets:insert(Tab,{integer_to_list(N),N}),
    ets:insert(Tab,{integer_to_list(N),N+1}),
    filltabstr2(Tab,N-1).
filltabstr3(Tab,0) ->
    Tab;
filltabstr3(Tab,N) ->
    ets:insert(Tab,{integer_to_list(N),N}),
    ets:insert(Tab,{integer_to_list(N),N}),
    filltabstr3(Tab,N-1).
xfilltabstr(Tab,N) ->
    case ets:info(Tab,type) of
	bag ->
	    filltabstr2(Tab,N);
	duplicate_bag ->
	    ets:select_delete(Tab,[{'_',[],[true]}]),
	    filltabstr3(Tab,N);
	_ ->
	    filltabstr(Tab,N)
    end.
fill_sets_int(N) ->
    Tab1 = ets:new(xxx, [ordered_set]),
    filltabint(Tab1,N),
    Tab2 = ets:new(xxx, [set]),
    filltabint(Tab2,N),
    Tab3 = ets:new(xxx, [bag]),
    filltabint2(Tab3,N),
    Tab4 = ets:new(xxx, [duplicate_bag]),
    filltabint3(Tab4,N),
    [Tab1,Tab2,Tab3,Tab4].

% 	       case ets:info(Table, type) of
% 		   bag ->
% 		       throw(Table);
% 		   _ ->
% 		       ok
% 	       end,
check_fun(Tab,Fun,'$end_of_table') ->
    ok;
check_fun(Tab,Fun,Item) ->
    lists:foreach(fun(Obj) ->
			  true = Fun(Obj)
		  end,
		  ets:lookup(Tab,Item)),
    check_fun(Tab,Fun,ets:next(Tab,Item)),
    ok.

check(Tab,Fun,N) ->
    N = ets:info(Tab, size),
    check_fun(Tab,Fun,ets:first(Tab)).



del_one_by_one_set(T,N,N) ->
    0 = ets:info(T,size),
    ok;
del_one_by_one_set(T,From,To) ->
    N = ets:info(T,size),
    ets:delete_object(T,{From, integer_to_list(From)}),
    N = (ets:info(T,size) + 1),
    Next = if
	       From < To ->
		   From + 1;
	       true ->
		   From - 1
	   end,
    del_one_by_one_set(T,Next,To).

del_one_by_one_bag(T,N,N) ->
    0 = ets:info(T,size),
    ok;
del_one_by_one_bag(T,From,To) ->
    N = ets:info(T,size),
    ets:delete_object(T,{From + From rem 2, integer_to_list(From)}),
    N = (ets:info(T,size) + 1),
    Next = if
	       From < To ->
		   From + 1;
	       true ->
		   From - 1
	   end,
    del_one_by_one_bag(T,Next,To).


del_one_by_one_dbag_1(T,N,N) ->
    0 = ets:info(T,size),
    ok;
del_one_by_one_dbag_1(T,From,To) ->
    N = ets:info(T,size),
    ets:delete_object(T,{From, integer_to_list(From)}),
    case From rem 2 of
	0 ->
	    N = (ets:info(T,size) + 2);
	1 ->
	    N = ets:info(T,size)
    end,
    Next = if
	       From < To ->
		   From + 1;
	       true ->
		   From - 1
	   end,
    del_one_by_one_dbag_1(T,Next,To).

del_one_by_one_dbag_2(T,N,N) ->
    0 = ets:info(T,size),
    ok;
del_one_by_one_dbag_2(T,From,To) ->
    N = ets:info(T,size),
    ets:delete_object(T,{From, integer_to_list(From)}),
    case From rem 2 of
	0 ->
	    N = (ets:info(T,size) + 3);
	1 ->
	    N = (ets:info(T,size) + 1)
    end,
    Next = if
	       From < To ->
		   From + 1;
	       true ->
		   From - 1
	   end,
    del_one_by_one_dbag_2(T,Next,To).

successive_delete(Table,From,To,Type) ->
    successive_delete(Table,From,To,Type,ets:info(Table,type)).

successive_delete(Table,N,N,_,_) ->
    ok;
successive_delete(Table,From,To,Type,TType) ->
    MS = case Type of
	     bound ->
		 [{{From,'_'},[],[true]}];
	     unbound ->
		 [{{'$1','_'},[],[{'==', '$1', From}]}]
	 end,
    case TType of
	X when X == bag; X == duplicate_bag ->
	    %erlang:display(From),
	    case From rem 2 of
		0 ->
		    2 = ets:select_delete(Table,MS);
		_ ->
		    0 = ets:select_delete(Table,MS)
	    end;
	_ ->
	    1 = ets:select_delete(Table,MS)
    end,
    Next = if
	       From < To ->
		   From + 1;
	       true ->
		   From - 1
	   end,
    successive_delete(Table, Next, To, Type,TType).

gen_dets_filename(Config,N) ->
    filename:join(?config(priv_dir,Config),
		  "testdets_" ++ integer_to_list(N) ++ ".dets").



