-module(id_transform_SUITE).
-author('pan@erix.ericsson.se').

-include_lib("kernel/include/file.hrl").
          
-compile(export_all).

% Serves as test...
-hej(hopp).
-include("test_server.hrl").

all(suite) -> [id_transform].

id_transform(doc) -> "Test erl_id_trans.";
id_transform(Config) when list(Config) ->
    ?line File=filename:join([code:lib_dir(stdlib),"examples",
			      "erl_id_trans.erl"]),
    ?line {ok,erl_id_trans,Bin}=compile:file(File,[binary]),
    ?line {module,erl_id_trans}=code:load_binary(erl_id_trans,File,Bin),
    ?line run_in_test_suite().

run_in_test_suite() ->
    LibDir = code:lib_dir(),
    SuperDir = 
	filename:dirname(filename:dirname(code:which(?MODULE))),
    TestDirs = filelib:wildcard(filename:join([SuperDir,"*_test"])),
    {All, Res} = case LibDir of
		     "/clearcase/otp/erts/lib" ->
			 %% Only test_suites cause clearcase is to slow...
			 {false, run_list(TestDirs,false)};
		     _ ->
			 {true, run_codepath_and(TestDirs)}
		 end,
    Comment0 = case All of
		   true ->
		       [];
		   false ->
		       "Only testsuite directories traversed, "
	       end,
    io:format("Result:~n~p~n", [Res]),
    case Res of
	{error, Reason} ->
	    {failed, Comment0 ++ lists:flatten(io_lib:format("~w",[Reason]))};
	{ok, RL0} ->
	    RL = lists:map(fun({Key,Value}) when is_list(Value) ->
				   {Key,length(Value)};
			      (Other) ->
				   Other
			   end,
			   RL0),
	    {comment,  Comment0 ++ lists:flatten(io_lib:format("~w",[RL]))}
    end.

run(Path) ->
    run_list([Path],true).

run_codepath() ->
    run_codepath_and([]).

run_codepath_and(DirList) ->
    ParentDirs = [filename:dirname(filename:absname(X)) || 
		     X <- code:get_path()],
    LibDir = code:lib_dir(),
    run_list([X || X <- ParentDirs, lists:prefix(X,LibDir) =:= false] ++ 
	     DirList,true).
		    

run_list(PathL,LibDirIncludes) ->
    io:format("Searching for header files...~n",[]),
    IP = build_ip(PathL,LibDirIncludes),
    io:format("...done~n",[]),
    io:format("Searching for erlang files ...~n",[]),
    Erls = collect_erls(PathL,[]),
    io:format("...done~n",[]),
    io:format("Applying erl_id_trans to found erlang files...~n",[]),
    Res = [do_trans(X,IP) || X <- Erls],
    io:format("...done~n",[]),
    Successes = [X || {ok,X} <- Res],
    LintFailures = [{F,E} || {failed,{F,{lint,E}}} <- Res],
    MnemosyneFailures = [F || {failed,{F,{mnemosyne,_E}}} <- Res],
    ParseFailures = [F || {failed,{F,{parse,_}}} <- Res],
    SevereFailures = [{F,E} || {failed,{F,{transform,E}}} <- Res],
    Stat = [{total_processed,length(Res)}, {success, length(Successes)},
	    {lint_failures, LintFailures},{parse_failures, ParseFailures},
	    {mnemosyne_failures, MnemosyneFailures}],
    case length(SevereFailures) of
	0 ->
	    {ok,Stat};
	_Else ->
	    {error, Stat ++ [{parse_transform_failed, SevereFailures}]}
    end.
    

collect_erls([],Accu) ->
    Accu;
collect_erls([Path0 | PathT],Accu) ->
    Path = filename:join([X || X <- filename:split(filename:absname(Path0)), 
			       X =/= "."]),    
    io:format("...in ~s~n",[Path]),
    Erls = find_files(Path,".erl"),
    collect_erls(PathT,Erls ++ Accu).
		     

build_ip(PathList,LibDirIncludes) ->
    LibDir0 = code:lib_dir(),
    {LibDir,Headers} = if
			   LibDirIncludes ->
			       io:format("...in ~p~n",[LibDir0]),
			       {LibDir0,find_files(LibDir0,".hrl")};
			   true ->
			       {false,[]}
		       end,
    DDirs = [filename:dirname(X) || X <- Headers],
    MoreDirs = build_ip(PathList,[],LibDir),
    lists:usort(DDirs ++ MoreDirs).

build_ip([],Accu,_LibDir) ->
    Accu;
build_ip([Path0|PathT],Accu,LibDir) ->
    Path = filename:join([X || X <- filename:split(filename:absname(Path0)), 
			     X =/= "."]),

    case lists:prefix(LibDir,Path) of
	true ->
	    build_ip(PathT,Accu,LibDir);
	false ->
	    io:format("...in ~p~n",[Path]),
	    Headers = find_files(Path,".hrl"),
	    DDirs = [filename:dirname(X) || X <- Headers],
	    NAccu = lists:usort(DDirs) ++ Accu,
	    build_ip(PathT,NAccu,LibDir)
    end.
    


% 	    filename:join([Dir,"..","..","kernel","include"]),
% 	    filename:join([Dir,"..","..","stdlib","include"]),
% 	    filename:join([Dir,"..","..","test_server","include"]),
% 	    filename:join([Dir,"..","..","hipe","main"]) | IncludePath],

do_trans(File,IncludePath) ->
    % Heuristic includePath...
    Dir = filename:dirname(File),
    Inc0 = [Dir, 
	    filename:join([Dir,"..","include"]),
	    filename:join([Dir,"..","..","..","src"]),
	    filename:join([Dir,"..","test_server"]),
 	    filename:join([Dir,"..","..","test_server","include"]),
	    filename:join([code:lib_dir(kernel),"include"]),
	    filename:join([code:lib_dir(stdlib),"include"])
	    | IncludePath],
    Inc = case code:lib_dir(hipe) of
	      {error,_} ->
		  Inc0;
	      HipeDir ->
		  Inc0 ++ [filename:join([HipeDir,"main"])]
	  end,
    % Defines to make various OTP applications checkable
    Defs = [{erlang_daemon_port,0},{epmd_dist_high,0}, {epmd_dist_low,0},
	    {inets_priv_dir,"."},{inets_data_dir,"."},
	    {'COMPILER_VSN',"0.0"},
	    {'TIPC_DIST_SERVER',"x"}, {'TIPC_DIST_RESOLVER',"y"},
	    {'TIPC_EPMD',"z"},{'APP',"tipc"},{'VSN',"0.0"},
	    {'TIPC_SERVER', "a"}, 
	    {'DSN', "0.0"},{'UID',0},{'PWD',"."},
	    {'ConnectStr', "x"},%{vsn,"0.0"}, breaks to much code
	    {'ORBVSN',"0.0"}],
    case (catch epp:parse_file(File,Inc,Defs)) of
	{ok, Tree0} ->
	    case mnemosyne_lc:parse_transform(Tree0,[]) of
		Tree when list(Tree) ->
		    case erl_lint:module(Tree) of
			{ok,_} ->
			    case (catch erl_id_trans:parse_transform(Tree0, 
								     [])) of
				Tree2 when list(Tree2) ->
				    {ok, File};
				{'EXIT', Reason} ->
				    {failed, {File, 
					      {transform,{crash, Reason}}}};
				Else ->
				    {failed, {File, 
					      {transform,{unknown, Else}}}}
			    end;
			
			Else0 ->
			    {failed, {File, {lint, Else0}}}
		    end;
		MnemError ->
		    {failed, {File, {mnemosyne,{unknown, MnemError}}}}
	    end;
	{'EXIT', Reason1} ->
	    {failed, {File, {parse,{crash, Reason1}}}};
	{error, Reason2} ->
	    {failed, {File, {parse,{error, Reason2}}}};
	Else1 ->
	    {failed, {File, {parse,{unknown, Else1}}}}
    end.

find_files(Path,Pattern) ->
    P2 = filename:join([X || X <- filename:split(filename:absname(Path)), 
			     X =/= "."]),
    find_files(P2,Pattern,[]).

find_files(Path,Pattern,Acc) ->
    case file:list_dir(Path) of
	{ok, L} ->
 	    All = [filename:join(Path,X) || X <- L], 
	    Matching = [X || X <- All, ends_with(X,Pattern)],
	    NewAcc1 = Matching ++ Acc,
	    Dirs = [X || X <- All, is_dir(X)],
	    find_in_dirs(Dirs,Pattern,NewAcc1);
	_ ->
	    Acc
    end.
	    

ends_with(Name,Pattern) ->
    LN = length(Name),
    LP = length(Pattern),
    if 
	LN < LP ->
	    false;
	true ->
	    Last = lists:sublist(Name,LN - LP + 1,LP),
	    Pattern =:= Last
    end.

is_dir(F) ->
    case file:read_file_info(F) of
	{ok,#file_info{type=directory}} ->
	    true;
	_ ->
	    false
    end.

find_in_dirs([],_,Acc) ->
    Acc;
find_in_dirs([Dir|T],Pattern,Acc) ->
    NAcc = find_files(Dir,Pattern,Acc),
    find_in_dirs(T,Pattern,NAcc).



% From here on there's only fake code to serve as test cases 
% for the id_transform.
% They need to be exported.

check(X,_Y) when X ->   
    true;
check(A,_) when atom(A) ->
    atom;
check(A,_) when erlang:is_list(A) ->
    list;
check(A,B) when erlang:'+'(A,B) ->
    atom;
check(_,_) ->
    false.

check2(A) ->
    case A of
	"hej" ++ "hopp" ->
	    a;
	[$h,$e,$j] ++ "hopp" ->
	    a;
	[1] ++ [2] ->
	    b
    end.

-record(x,{x,y}).
-record(y,{x=1,y=0}).

g() ->
    #y.y.

f(#y.y) ->
    vansinne;

f(X) when X =:= #y.y ->
    {#y.y,_Y} = {X,hej};
f(X) ->
    if X#x.y ->
	    ok;
       element(3,X) ->
	    banan;
       true ->
	    nok
    end;

f(#x{_='_'}) ->
    hopp;
f(#x{x=true,y=true}) ->
    babba;
f(A) when A == #x{x=true,y=true} ->
    true;
f(A) when A#x.x == 4 ->
    #x{x = 1, _ = 2}.

% Stolen from erl_lint_SUITE.erl
-record(apa, {}).

t(A) when atom(A) ->
    atom;
t(A) when binary(A) ->
    binary;
t(A) when constant(A) ->
    constant;
t(A) when float(A) ->
    float;
t(A) when function(A) ->
    function;
t(A) when integer(A) ->
    integer;
t(A) when is_atom(A) ->
    is_atom;
t(A) when is_binary(A) ->
    is_binary;
t(A) when is_constant(A) ->
    is_constant;
t(A) when is_float(A) ->
    is_float;
t(A) when is_function(A) ->
    is_function;
t(A) when is_integer(A) ->
    is_integer;
t(A) when is_list(A) ->
    is_list;
t(A) when is_number(A) ->
    is_number;
t(A) when is_pid(A) ->
    is_pid;
t(A) when is_port(A) ->
    is_port;
t(A) when is_record(A, apa) ->
    is_record;
t(A) when is_reference(A) ->
    is_reference;
t(A) when is_tuple(A) ->
    is_tuple;
t(A) when list(A) ->
    list;
t(A) when number(A) ->
    number;
t(A) when pid(A) ->
    pid;
t(A) when port(A) ->
    port;
t(A) when record(A, apa) ->
    record;
t(A) when reference(A) ->
    reference;
t(A) when tuple(A) ->
    tuple.

t1(A) when atom(A), atom(A) ->
    atom;
t1(A) when binary(A), binary(A) ->
    binary;
t1(A) when constant(A), constant(A) ->
    constant;
t1(A) when float(A), float(A) ->
    float;
t1(A) when function(A), function(A) ->
    function;
t1(A) when integer(A), integer(A) ->
    integer;
t1(A) when is_atom(A), is_atom(A) ->
    is_atom;
t1(A) when is_binary(A), is_binary(A) ->
    is_binary;
t1(A) when is_constant(A), is_constant(A) ->
    is_constant;
t1(A) when is_float(A), is_float(A) ->
    is_float;
t1(A) when is_function(A), is_function(A) ->
    is_function;
t1(A) when is_integer(A), is_integer(A) ->
    is_integer;
t1(A) when is_list(A), is_list(A) ->
    is_list;
t1(A) when is_number(A), is_number(A) ->
    is_number;
t1(A) when is_pid(A), is_pid(A) ->
    is_pid;
t1(A) when is_port(A), is_port(A) ->
    is_port;
t1(A) when is_record(A, apa), is_record(A, apa) ->
    is_record;
t1(A) when is_reference(A), is_reference(A) ->
    is_reference;
t1(A) when is_tuple(A), is_tuple(A) ->
    is_tuple;
t1(A) when list(A), list(A) ->
    list;
t1(A) when number(A), number(A) ->
    number;
t1(A) when pid(A), pid(A) ->
    pid;
t1(A) when port(A), port(A) ->
    port;
t1(A) when record(A, apa), record(A, apa) ->
    record;
t1(A) when reference(A), reference(A) ->
    reference;
t1(A) when tuple(A), tuple(A) ->
    tuple.

t2(A) when atom(A); atom(A) ->
    atom;
t2(A) when binary(A); binary(A) ->
    binary;
t2(A) when constant(A); constant(A) ->
    constant;
t2(A) when float(A); float(A) ->
    float;
t2(A) when function(A); function(A) ->
    function;
t2(A) when integer(A); integer(A) ->
    integer;
t2(A) when is_atom(A); is_atom(A) ->
    is_atom;
t2(A) when is_binary(A); is_binary(A) ->
    is_binary;
t2(A) when is_constant(A); is_constant(A) ->
    is_constant;
t2(A) when is_float(A); is_float(A) ->
    is_float;
t2(A) when is_function(A); is_function(A) ->
    is_function;
t2(A) when is_integer(A); is_integer(A) ->
    is_integer;
t2(A) when is_list(A); is_list(A) ->
    is_list;
t2(A) when is_number(A); is_number(A) ->
    is_number;
t2(A) when is_pid(A); is_pid(A) ->
    is_pid;
t2(A) when is_port(A); is_port(A) ->
    is_port;
t2(A) when is_record(A, apa); is_record(A, apa) ->
    is_record;
t2(A) when is_reference(A); is_reference(A) ->
    is_reference;
t2(A) when is_tuple(A); is_tuple(A) ->
    is_tuple;
t2(A) when list(A); list(A) ->
    list;
t2(A) when number(A); number(A) ->
    number;
t2(A) when pid(A); pid(A) ->
    pid;
t2(A) when port(A); port(A) ->
    port;
t2(A) when record(A, apa); record(A, apa) ->
    record;
t2(A) when reference(A); reference(A) ->
    reference;
t2(A) when tuple(A); tuple(A) ->
    tuple.

t3(A) when is_atom(A) or is_atom(A) ->
    is_atom;
t3(A) when is_binary(A) or is_binary(A) ->
    is_binary;
t3(A) when is_constant(A) or is_constant(A) ->
    is_constant;
t3(A) when is_float(A) or is_float(A) ->
    is_float;
t3(A) when is_function(A) or is_function(A) ->
    is_function;
t3(A) when is_integer(A) or is_integer(A) ->
    is_integer;
t3(A) when is_list(A) or is_list(A) ->
    is_list;
t3(A) when is_number(A) or is_number(A) ->
    is_number;
t3(A) when is_pid(A) or is_pid(A) ->
    is_pid;
t3(A) when is_port(A) or is_port(A) ->
    is_port;
t3(A) when is_record(A, apa) or is_record(A, apa) ->
   is_record;
t3(A) when is_reference(A) or is_reference(A) ->
   is_reference;
t3(A) when is_tuple(A) or is_tuple(A) ->
   is_tuple; 
t3(A) when record(A, apa) ->
   foo;
% Compiler crash 
%t3(A) when {erlang,is_record}(A, apa) ->
%   foo;
t3(A) when erlang:is_record(A, apa) ->
   foo;
t3(A) when is_record(A, apa) ->
   foo;
t3(A) when record({apa}, apa) ->
   {A,foo};
% Compiler crash 
%t3(A) when {erlang,is_record}({apa}, apa) ->
%   foo;
t3(A) when erlang:is_record({apa}, apa) ->
   {A,foo};
t3(A) when is_record({apa}, apa) ->
   {A,foo}.

-record(apa2,{a=a,b=foo:bar()}).
apa() ->
    [X || X <- [], #apa2{a = a} == {r,X,foo}];
apa() ->
    [X || X <- [], #apa2{b = b} == {r,X,foo}];
apa() ->
    [X || X <- [], 3 == X#apa2.a].






