%% ``The contents of this file are subject to the Erlang Public License,
%% Version 1.1, (the "License"); you may not use this file except in
%% compliance with the License. You should have received a copy of the
%% Erlang Public License along with this software. If not, it can be
%% retrieved via the world wide web at http://www.erlang.org/.
%% 
%% Software distributed under the License is distributed on an "AS IS"
%% basis, WITHOUT WARRANTY OF ANY KIND, either express or implied. See
%% the License for the specific language governing rights and limitations
%% under the License.
%% 
%% The Initial Developer of the Original Code is Ericsson Utvecklings AB.
%% Portions created by Ericsson are Copyright 1999, Ericsson Utvecklings
%% AB. All Rights Reserved.''
%% 
%%     $Id$
%%
-module(queue_SUITE).
-export([all/1]).

-export([do/1, to_list/1, io_test/1, op_test/1, error/1]).

-export([init_per_testcase/2, fin_per_testcase/2]).

-include("test_server.hrl").

% Default timetrap timeout (set in init_per_testcase).
-define(default_timeout, ?t:minutes(1)).

init_per_testcase(Case, Config) ->
    ?line Dog = ?t:timetrap(?default_timeout),
    [{watchdog, Dog} | Config].
fin_per_testcase(Case, Config) ->
    Dog = ?config(watchdog, Config),
    test_server:timetrap_cancel(Dog),
    ok.

all(doc) ->
    ["Test cases for queue."];
all(suite) ->
    [do, to_list, io_test, op_test, error].

do(doc) ->
    [""];
do(suite) ->
    [];
do(Config) when list(Config) ->
    ?line E = queue:new(),
    ?line [] = queue:to_list(E),

    ?line do_queue(E,
		   [{in, 1},
		    {in, 2},
		    {out, {value, 1}},
		    {in, 3},
		    {out, {value, 2}},
		    {out, {value, 3}},
		    {out, empty}
		   ]),
    ok.

to_list(doc) ->
    ["OTP-2701"];
to_list(suite) ->
    [];
to_list(Config) when list(Config) ->
    ?line E = queue:new(),
    ?line Q = do_queue(E, [{in, 1},
			   {in, 2},
			   {in, 3},
			   {out, {value, 1}},
			   {in, 4},
			   {in, 5}]),
    ?line case queue:to_list(Q) of
	      [2,3,4,5] ->
		  ok;
	      Other ->
		  test_server:fail(Other)
	  end,

    ok.

do_queue(Q, []) ->
    Q;
do_queue(Q, [E | Rest]) ->
    do_queue(do_queue_1(E, Q), Rest).

do_queue_1({in, E}, Q) ->
    queue:in(E, Q);
do_queue_1({out, E}, Q) ->
    case queue:out(Q) of
	{E, Q1} ->
	    Q1;
	Other ->
	    test_server:fail({"out failed", E, Q, Other})
    end.


io_test(doc) ->
    "Test input and output";
io_test(suite) ->
    [];
io_test(Config) when list(Config) ->
    ?line [4,3,5] = 
	io([snoc,snoc,head,head,head,cons,cons,snoc], queue:new(), 1),
    ?line [5,3,4] = 
	io([cons,cons,daeh,daeh,daeh,snoc,snoc,cons], queue:new(), 1),
    ?line [4,3,5] = 
	io([in,in,out,out,out,in_r,in_r,in], queue:new(), 1),
    ?line [5,3,4] = 
	io([in_r,in_r,out_r,out_r,out_r,in,in,in_r], queue:new(), 1),
    %%
    ?line [] = 
	io([snoc,snoc,head,snoc,snoc,head,head,snoc,head,head], queue:new(), 1),
    ?line [] = 
	io([cons,cons,daeh,cons,cons,daeh,daeh,cons,daeh,daeh], queue:new(), 1),
    ?line [] = 
	io([in,in,out,in,in,out,out,in,out,out], queue:new(), 1),
    ?line [] = 
	io([in_r,in_r,out_r,in_r,in_r,out_r,out_r,in_r,out_r,out_r], 
	   queue:new(), 1),
    %%
    ?line [5,6] = 
	io([snoc,snoc,snoc,head,head,snoc,snoc,snoc,head,head], queue:new(), 1),
    ?line [6,5] = 
	io([cons,cons,cons,daeh,daeh,cons,cons,cons,daeh,daeh], queue:new(), 1),
    ?line [5,6] = 
	io([in,in,in,out,out,in,in,in,out,out], queue:new(), 1),
    ?line [6,5] = 
	io([in_r,in_r,in_r,out_r,out_r,in_r,in_r,in_r,out_r,out_r], 
	   queue:new(), 1),
    %%
    ?line [5] = 
	io([snoc,head,head,snoc,head,snoc,head,snoc,head,snoc], queue:new(), 1),
    ?line [5] = 
	io([cons,daeh,daeh,cons,daeh,cons,daeh,cons,daeh,cons], queue:new(), 1),
    ?line [5] = 
	io([in,out,out,in,out,in,out,in,out,in], queue:new(), 1),
    ?line [5] = 
	io([in_r,out_r,out_r,in_r,out_r,in_r,out_r,in_r,out_r,in_r], 
	   queue:new(), 1),
    %%
    ?line [] =
	io([snoc,head,snoc,snoc,head,head,snoc,snoc,snoc,head,head,head], 
	   queue:new(), 1),
    ?line [] =
	io([cons,daeh,cons,cons,daeh,daeh,cons,cons,cons,daeh,daeh,daeh],
	  queue:new(), 1),
    ?line [] =
	io([in,out,in,in,out,out,in,in,in,out,out,out], 
	   queue:new(), 1),
    ?line [] =
	io([in_r,out_r,in_r,in_r,out_r,out_r,in_r,in_r,in_r,out_r,out_r,out_r],
	  queue:new(), 1),
    %%
    ?line [3] =	io([cons,cons,cons,snoc,daeh,daeh,daeh], queue:new(), 1),
    ?line [3] =	io([snoc,snoc,snoc,cons,head,head,head], queue:new(), 1),
    ?line [3] =	io([in,in,in,in_r,out,out,out], queue:new(), 1),
    ?line [3] =	io([in_r,in_r,in_r,in,out_r,out_r,out_r], queue:new(), 1),
    %%
    Q2 = queue:join(queue:cons(1, queue:new()),queue:cons(2, queue:new())),
    Q1 = queue:reverse(Q2),
    ?line [1] = io([head],  Q1, 3),
    ?line [1] = io([out],   Q1, 3),
    ?line [1] = io([daeh],  Q2, 3),
    ?line [1] = io([out_r], Q2, 3),
%    ?line [2] =	io([cons,cons,snoc,daeh,daeh], [], 1),
%    ?line [2] =	io([snoc,snoc,cons,head,head], [], 1),
%    ?line [2] =	io([in,in,in_r,out,out], [], 1),
%    ?line [2] =	io([in_r,in_r,in,out_r,out_r], [], 1),
    %%
    ok.

%% Perform a list of operations to a queue.
%% Keep a reference queue on the side; just a list.
%% Compare the read values between the queues.
%% Return the resulting queue as a list.
%% Inserted values are increments of the previously inserted.
io(Ops, Q, X) ->
    io(Ops, Q, queue:to_list(Q), X).

io([head | Tail], Q, [], X) ->
    true = queue:is_empty(Q),
    {'EXIT',{empty,_}} = (catch {ok,queue:head(Q)}),
    {'EXIT',{empty,_}} = (catch {ok,queue:tail(Q)}),
    io(Tail, Q, [], X);
io([head | Tail], Q, [H | T], X) ->
    H = queue:head(Q),
    false = queue:is_empty(Q),
    io(Tail, queue:tail(Q), T, X);
io([daeh | Tail], Q, [], X) ->
    true = queue:is_empty(Q),
    {'EXIT',{empty,_}} = (catch {ok,queue:daeh(Q)}),
    {'EXIT',{empty,_}} = (catch {ok,queue:lait(Q)}),
    io(Tail, Q, [], X);
io([daeh | Tail], Q, QQ, X) ->
    H = queue:daeh(Q),
    false = queue:is_empty(Q),
    [H | T] = lists:reverse(QQ),
    io(Tail, queue:lait(Q), lists:reverse(T), X);
io([out | Tail], Q, [], X) ->
    {empty, Q1} = queue:out(Q),
    io(Tail, Q1, [], X);
io([out | Tail], Q, [H | T], X) ->
    {{value,H}, Q1} = queue:out(Q),
    io(Tail, Q1, T, X);
io([out_r | Tail], Q, [], X) ->
    {empty, Q1} = queue:out_r(Q),
    io(Tail, Q1, [], X);
io([out_r | Tail], Q, QQ, X) ->
    {{value,H}, Q1} = queue:out_r(Q),
    [H | T] = lists:reverse(QQ),
    io(Tail, Q1, lists:reverse(T), X);
io([cons | Tail], Q, QQ, X) ->
    io(Tail, queue:cons(X,Q), [X|QQ], X+1);
io([snoc | Tail], Q, QQ, X) ->
    io(Tail, queue:snoc(Q,X), QQ++[X], X+1);
io([in_r | Tail], Q, QQ, X) ->
    io(Tail, queue:in_r(X,Q), [X|QQ], X+1);
io([in | Tail], Q, QQ, X) ->
    io(Tail, queue:in(X,Q), QQ++[X], X+1);
io([], Q, QQ, _X) ->
    QQ = queue:to_list(Q),
    Length = length(QQ),
    Length = queue:len(Q),
    QQ.


op_test(doc) ->
    "Test operations on whole queues";
op_test(suite) ->
    [];
op_test(Config) when list(Config) ->
    ?line Len = 50,
    ?line Len2 = 2*Len,
    ?line L1 = lists:seq(1, Len),
    ?line L1r = lists:reverse(L1),
    ?line L2 = lists:seq(Len+1, Len2),
    ?line L2r = lists:reverse(L2),
    ?line L3 = L1++L2,
    ?line L3r = L2r++L1r,
    ?line Q0 = queue:new(),
    ?line [] = queue:to_list(Q0),
    ?line Q0 = queue:from_list([]),
    ?line Q1 = queue:from_list(L1),
    ?line Q2 = queue:from_list(L2),
    ?line Q3 = queue:from_list(L3),
    ?line Len = queue:len(Q1),
    ?line Len = queue:len(Q2),
    ?line Len2 = queue:len(Q3),
    ?line L1 = queue:to_list(Q1),
    ?line L2 = queue:to_list(Q2),
    ?line L3 = queue:to_list(Q3),
    ?line Q3b = queue:join(Q0, queue:join(queue:join(Q1, Q2), Q0)),
    ?line L3 = queue:to_list(Q3b),
    ?line {Q0, Q3} = queue:split(0, Q3),
    ?line {Q3, Q0} = queue:split(Len2, Q3),
    ?line {Q1a, Q2a} = queue:split(Len, Q3),
    ?line L1 = queue:to_list(Q1a),
    ?line L2 = queue:to_list(Q2a),
    ?line {Q1b, Q2b} = queue:split(Len, Q3b),
    ?line L1 = queue:to_list(Q1b),
    ?line L2 = queue:to_list(Q2b),
    ?line Len = queue:len(Q1b),
    ?line Len = queue:len(Q2b),
    ?line Len2 = queue:len(Q3b),
    ?line Q1r = queue:reverse(Q1),
    ?line Q2r = queue:reverse(Q2),
    ?line Q1ar = queue:reverse(Q1a),
    ?line Q2ar = queue:reverse(Q2a),
    ?line Q1br = queue:reverse(Q1b),
    ?line Q2br = queue:reverse(Q2b),
    ?line Q3br = queue:reverse(Q3b),
    ?line L1r = queue:to_list(Q1r),
    ?line L1r = queue:to_list(Q1ar),
    ?line L1r = queue:to_list(Q1br),
    ?line L2r = queue:to_list(Q2r),
    ?line L2r = queue:to_list(Q2ar),
    ?line L2r = queue:to_list(Q2br),
    ?line L3r = queue:to_list(Q3br),
    ?line Len = queue:len(Q1br),
    ?line Len = queue:len(Q2br),
    ?line Len2 = queue:len(Q3br),
    ok.

error(doc) ->
    "Test queue errors";
error(suite) ->
    [];
error(Config) when list(Config) ->
    ?line {'EXIT',{badarg,_}} = (catch queue:in(1, illegal_queue)),
    ?line {'EXIT',{badarg,_}} = (catch queue:out(illegal_queue)),
    ?line {'EXIT',{badarg,_}} = (catch queue:in_r(1, illegal_queue)),
    ?line {'EXIT',{badarg,_}} = (catch queue:out_r(illegal_queue)),
    ?line {'EXIT',{badarg,_}} = (catch queue:to_list(illegal_queue)),
    %%
    ?line {'EXIT',{badarg,_}} = (catch queue:from_list(no_list)),
    ?line {'EXIT',{badarg,_}} = (catch queue:is_empty(no_queue)),
    ?line {'EXIT',{badarg,_}} = (catch queue:len(no_queue)),
    %%
    ?line {'EXIT',{badarg,_}} = (catch queue:cons(1, no_queue)),
    ?line {'EXIT',{badarg,_}} = (catch queue:head(no_queue)),
    ?line {'EXIT',{badarg,_}} = (catch queue:tail(no_queue)),
    %%
    ?line {'EXIT',{badarg,_}} = (catch queue:snoc(no_queue, 1)),
    ?line {'EXIT',{badarg,_}} = (catch queue:last(no_queue)),
    ?line {'EXIT',{badarg,_}} = (catch queue:daeh(no_queue)),
    ?line {'EXIT',{badarg,_}} = (catch queue:lait(no_queue)),
    ?line {'EXIT',{badarg,_}} = (catch queue:init(no_queue)),
    %%
    ?line {'EXIT',{badarg,_}} = (catch queue:reverse(no_queue)),
    ?line {'EXIT',{badarg,_}} = (catch queue:join(queue:new(), no_queue)),
    ?line {'EXIT',{badarg,_}} = (catch queue:join(no_queue, queue:new())),
    ?line {'EXIT',{badarg,_}} = (catch queue:split(17, no_queue)),
    ?line {'EXIT',{badarg,_}} = (catch queue:head(no_queue)),
    %%
    ?line Q0 = queue:new(),
    ?line {'EXIT',{badarg,_}} = (catch queue:split(1, Q0)),
    ?line {'EXIT',{badarg,_}} = (catch queue:split(2, queue:snoc(Q0, 1))),
    ok.
